# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

import gi
gi.require_version('Hinawa', '3.0')
from gi.repository import Hinawa

from array import array
from math import log10

__all__ = ['EftInfo', 'EftFlash', 'EftTransmit', 'EftHwctl', 'EftPhysOutput',
           'EftPhysInput', 'EftPlayback', 'EftCapture', 'EftMonitor',
           'EftIoconf']

#
# Category No.0, for hardware information
#


class EftInfo():
    SUPPORTED_MODELS = (
        'Audiofire2',
        'Audiofire4',
        'Audiofire8',
        'Audiofire8p',
        'Audiofire12',
        'Audiofire12HD',
        'Audiofire12Apple',
        'FireworksHDMI',
        'Onyx400F',
        'Onyx1200f',
        'Fireworks8',
        'RobotInterfacePack',
        'AudioPunk',
    )

    SUPPORTED_FEATURES = (
        'changeable-resp-addr',
        'control-room-mirroring',
        'spdif-coax',
        'aesebu-xlr',
        'dsp',
        'fpga',
        'phantom-powering',
        'rx-mapping',
        'adjust-input-level',
        'spdif-opt',
        'adat-opt',
        'nominal-input',
        'nominal-output',
        'soft-clip',
        'robot-hex-input',
        'robot-battery-charge',
        # For our purpose
        'tx-mapping',
    )

    SUPPORTED_CLOCK_SOURCES = (
        'internal',
        'syt-match',
        'word-clock',
        'spdif',
        'adat1',
        'adat2',
        'continuous'
    )

    SUPPORTED_SAMPLING_RATES = (
        32000, 44100, 48000,
        88200, 96000,
        176400, 192000
    )

    SUPPORTED_FIRMWARES = ('ARM', 'DSP', 'FPGA')

    SUPPORTED_PORT_NAMES = (
        'analog',
        'spdif',
        'adat',
        'spdif/adat',
        'analog mirroring',
        'headphones',
        'I2S',
        'guitar',
        'piezo guitar',
        'guitar string'
    )

    __MODELS = {
        'Audiofire2':           0x000af2,
        'Audiofire4':           0x000af4,
        'Audiofire8':           0x000af8,
        'Audiofire8p':          0x000af9,
        'Audiofire12':          0x00af12,
        'Audiofire12HD':        0x0af12d,
        'Audiofire12Apple':     0x0af12a,
        'FireworksHDMI':        0x00afd1,
        'Onyx400F':             0x00400f,
        'Onyx1200f':            0x01200f,
        'Fireworks8':           0x0000f8,
        'RobotInterfacePack':   0x00afb2,
        'AudioPunk':            0x00afb9,
    }

    __FEATURE_FLAGS = {
        'changeable-resp-addr':    0x0001,
        'control-room-mirroring':  0x0002,
        'spdif-coax':              0x0004,
        'aesebu-xlr':              0x0008,
        'dsp':                     0x0010,
        'fpga':                    0x0020,
        'phantom-powering':        0x0040,
        'rx-mapping':              0x0080,
        'adjust-input-level':      0x0100,
        'spdif-opt':               0x0200,
        'adat-opt':                0x0400,
        'nominal-input':           0x0800,
        'nominal-output':          0x1000,
        'soft-clip':               0x2000,
        'robot-hex-input':         0x4000,
        'robot-battery-charge':    0x8000,
    }

    __CLOCK_FLAGS = {
        'internal':    0x0001,
        'syt-match':   0x0002,
        'word-clock':  0x0004,
        'spdif':       0x0008,
        'adat1':       0x0010,
        'adat2':       0x0020,
        'continuous':  0x0040,
    }

    __MIDI_FLAGS = {
        'midi-in-1':        0x00000100,
        'midi-out-1':       0x00000200,
        'midi-in-2':        0x00000400,
        'midi-out-2':       0x00000800,
    }

    __ROBOT_FLAGS = {
        'battery-charging': 0x20000000,
        'stereo-connect':   0x40000000,
        'hex-signal':       0x80000000,
    }

    @staticmethod
    def _execute_command(unit, cmd, args):
        if not isinstance(unit, Hinawa.SndEfw):
            raise ValueError('Invalid argument for SndEfw')
        params = [0] * 256
        return unit.transaction(0, cmd, args, params)

    @classmethod
    def get_spec(cls, unit):
        params = cls._execute_command(unit, 0, None)
        info = {}
        for model, value in cls.__MODELS.items():
            if value == params[3]:
                info['model'] = model
                break
        else:
            raise RuntimeError('This model is not supported')
        info['features'] = cls._parse_capability(params)
        info['clock-sources'] = cls._parse_clock_source(params)
        info['sampling-rates'] = cls._parse_sampling_rate(params)
        info.update(cls._parse_phys_ports(params))
        info.update(cls._parse_mixer_channels(params))
        info.update(cls._parse_stream_formation(params))
        info['firmware-versions'] = cls._parse_firmware_versions(params)
        return info

    @classmethod
    def get_metering(cls, unit):
        params = cls._execute_command(unit, 1, None)
        metering = {}
        metering['clocks'] = {}
        for name, flag in cls.__CLOCK_FLAGS.items():
            if params[0] & flag:
                metering['clocks'][name] = True
            else:
                metering['clocks'][name] = False
        metering['midi'] = {}
        for name, flag in cls.__MIDI_FLAGS.items():
            if params[0] & flag:
                metering['midi'][name] = True
            else:
                metering['midi'][name] = False
        metering['robot'] = {}
        for name, flag in cls.__ROBOT_FLAGS.items():
            if params[0] & flag:
                metering['robot'][name] = True
            else:
                metering['robot'][name] = False
        metering['spdif'] = params[1]
        metering['adat'] = params[2]
        metering['outputs'] = []
        for i in range(params[5]):
            index = 9 + i
            if params[index] == 0:
                db = -144.0
            else:
                db = round(20 * log10(params[index] / 0x80000000), 1)
            metering['outputs'].append(db)
        metering['inputs'] = []
        for i in range(params[6]):
            index = 9 + params[5] + i
            if params[index] == 0:
                db = -144.0
            else:
                db = round(20 * log10(params[index] / 0x80000000), 1)
            metering['inputs'].append(db)
        return metering

    @classmethod
    def set_resp_addr(cls, unit, addr):
        args = array('I')
        args.append((addr >> 24) & 0xffffffff)
        args.append(addr & 0xffffffff)
        cls._execute_command(unit, 2, args)

    # 64 quads can be read at once.
    @classmethod
    def read_session_data(cls, unit, offset, quadlets):
        args = array('I')
        args.append(offset)
        args.append(quadlets)
        params = cls._execute_command(unit, 3, args)
        return params

    @classmethod
    def get_debug_info(cls, unit):
        params = cls._execute_command(unit, 4, None)

        # params[00]: isochronous stream 1 flushed
        # params[01]: isochronous stream 1 underruns
        # params[02]: firewire3 control
        # params[03]: firewire3 control written
        # params[04-15]: data
        return params

    @classmethod
    def test_dsp(cls, unit, value):
        args = array('I')
        args.append(value)
        params = cls._execute_command(unit, 5, args)
        return params[0]

    @classmethod
    def test_arm(cls, unit, value):
        args = array('I')
        args.append(value)
        params = cls._execute_command(unit, 6, args)
        return params[0]

    @classmethod
    def _parse_capability(cls, params):
        caps = {}
        for name, flag in cls.__FEATURE_FLAGS.items():
            if params[0] & flag:
                caps[name] = True
            else:
                caps[name] = False
        return caps

    @classmethod
    def _parse_clock_source(cls, params):
        srcs = {}
        for name, flag in cls.__CLOCK_FLAGS.items():
            if params[21] & flag:
                srcs[name] = True
            else:
                srcs[name] = False
        return srcs

    @classmethod
    def _parse_sampling_rate(cls, params):
        rates = {}
        for rate in cls.SUPPORTED_SAMPLING_RATES:
            if params[39] <= rate and rate <= params[38]:
                rates[rate] = True
            else:
                rates[rate] = False
        return rates

    @classmethod
    def _parse_phys_ports(cls, params):
        def parse_ports(params):
            ports = []
            data = (params[1] >> 16, params[1] & 0xffff,
                    params[2] >> 16, params[2] & 0xffff,
                    params[3] >> 16, params[3] & 0xffff,
                    params[4] >> 16, params[4] & 0xffff)
            for i in range(params[0]):
                count = data[i] & 0xff
                index = data[i] >> 8
                if index > len(cls.SUPPORTED_PORT_NAMES):
                    name = 'dummy'
                else:
                    name = cls.SUPPORTED_PORT_NAMES[index]
                for j in range(count):
                    ports.append(name)
            return ports

        return {'phys-inputs': parse_ports(params[31:]),
                'phys-outputs': parse_ports(params[26:])}

    @staticmethod
    def _parse_mixer_channels(params):
        return {'capture-channels': params[43],
                'playback-channels': params[42]}

    @staticmethod
    def _parse_stream_formation(params):
        return {'tx-stream-channels': (params[23], params[46], params[48]),
                'rx-stream-channels': (params[22], params[45], params[47])}

    @classmethod
    def _parse_firmware_versions(cls, params):
        return {'DSP':  cls._get_literal_version(params[40]),
                'ARM':  cls._get_literal_version(params[41]),
                'FPGA': cls._get_literal_version(params[44])}

    @staticmethod
    def _get_literal_version(val):
        return '{0}.{1}.{2}'.format((val >> 24) & 0xff,
                                    (val >> 16) & 0xff,
                                    (val >> 8) & 0xff)

#
# Category No.1, for flash commands
#


class EftFlash():
    @staticmethod
    def _execute_command(unit, cmd, args):
        if not isinstance(unit, Hinawa.SndEfw):
            raise ValueError('Invalid argument for SndEfw')
        params = [0] * 256
        return unit.transaction(1, cmd, args, params)

    @classmethod
    def erase(cls, unit, offset):
        args = array('I')
        args.append(offset)
        cls._execute_command(unit, 0, args)

    @classmethod
    def read_block(cls, unit, offset, quadlets):
        args = array('I')
        args.append(offset)
        args.append(quadlets)
        resp = cls._execute_command(unit, 1, args)
        if resp[0] != offset:
            raise OSError('Unexpected parameter for offset in response.')
        if resp[1] != quadlets:
            raise OSError('Unexpected parameter for quadlets in response.')
        return resp[2:]

    @classmethod
    def write_block(cls, unit, offset, data):
        args = array('I')
        args.append(offset)
        args.append(len(data))
        for datum in data:
            args.append(datum)
        cls._execute_command(unit, 2, args)

    @classmethod
    def get_status(cls, unit):
        # return status means it.
        cls._execute_command(unit, 3, None)

    @classmethod
    def get_session_offset(cls, unit):
        params = cls._execute_command(unit, 4, None)
        return params[0]

    @classmethod
    def set_lock(cls, unit, lock):
        args = array('I')
        if lock > 0:
            args.append(1)
        else:
            args.append(0)
        cls._execute_command(unit, 5, args)

#
# Category No.2, for transmission control commands
#


class EftTransmit():
    SUPPORTED_MODES = ('windows', 'iec61883-6')
    SUPPORTED_PLAYBACK_DROPS = (1, 2, 4)
    SUPPORTED_RECORD_STREATCH_RATIOS = (1, 2, 4)
    SUPPORTED_SERIAL_BPS = (16, 24)
    SUPPORTED_SERIAL_DATA_FORMATS = ('left-adjusted', 'i2s')

    @staticmethod
    def _execute_command(unit, cmd, args):
        if not isinstance(unit, Hinawa.SndEfw):
            raise ValueError('Invalid argument for SndEfw')
        params = [0] * 256
        return unit.transaction(2, cmd, args, params)

    @classmethod
    def set_mode(cls, unit, mode):
        if mode not in cls.SUPPORTED_MODES:
            raise ValueError('Invalid argument for mode')
        args = array('I')
        args.append(cls.SUPPORTED_MODES.index(mode))
        cls._execute_command(unit, 0, args)

    @classmethod
    def set_fw_hdmi(cls, unit, playback_drop, record_stretch_ratio, serial_bps,
                    serial_data_format):
        if playback_drop not in cls.SUPPORTED_PLAYBACK_DROPS:
            raise ValueError('Invalid argument for playback drop')
        if cls.SUPPORTED_RECORD_STREATCH_RATIOS(record_stretch_ratio) == 0:
            raise ValueError('Invalid argument for record stretch')
        if serial_bps not in cls.SUPPORTED_SERIAL_BPS:
            raise ValueError('Invalid argument for serial bits per second')
        if cls.SUPPORTED_SERIAL_DATA_FORMATS(serial_data_format) == 0:
            raise ValueError('Invalid argument for serial data format')

        args = array('I')
        args.append(playback_drop)
        args.append(record_stretch_ratio)
        args.append(serial_bps)
        args.append(cls.SUPPORTED_SERIAL_DATA_FORMATS.index(
            serial_data_format))
        cls._execute_command(unit, 4, args)

#
# Category No.3, for hardware control commands
#


class EftHwctl():
    SUPPORTED_BOX_STATES = {
        # name                  clear           set
        'internal-multiplexer': ('Disabled',    'Enabled'),
        'spdif-pro':            ('Disabled',    'Enabled'),
        'spdif-non-audio':      ('Disabled',    'Enabled'),
        'control-room':         ('A',           'B'),
        'output-level-bypass':  ('Disabled',    'Enabled'),
        'metering-mode-in':     ('A',           'B'),
        'metering-mode-out':    ('D1',          'D2'),
        'soft-clip':            ('Disabled',    'Enabled'),
        'robot-hex-input':      ('Disabled',    'Enabled'),
        'robot-battery-charge': ('Disabled',    'Enabled'),
        'phantom-powering':     ('Disabled',    'Enabled'),
    }

    # Internal parameters
    __BOX_STATE_POSITIONS = {
        # identifier            shift
        'internal-multiplexer':  0,
        'spdif-pro':             1,
        'spdif-non-audio':       2,
        'control-room':          8,
        'output-level-bypass':   9,
        'metering-mode-in':     12,
        'metering-mode-out':    13,
        'soft-clip':            16,
        'robot-hex-input':      29,
        'robot-battery-charge': 30,
        'phantom-powering':     31,
    }

    @staticmethod
    def _execute_command(unit, cmd, args):
        if not isinstance(unit, Hinawa.SndEfw):
            raise ValueError('Invalid argument for SndEfw')
        params = [0] * 256
        return unit.transaction(3, cmd, args, params)

    @classmethod
    def set_clock(cls, unit, rate, source, reset):
        if rate not in EftInfo.SUPPORTED_SAMPLING_RATES:
            raise ValueError('Invalid argument for sampling rate')
        if source not in EftInfo.SUPPORTED_CLOCK_SOURCES:
            raise ValueError('Invalid argument for source of clock')
        if reset > 0:
            reset = 0x80000000
        args = array('I')
        args.append(EftInfo.SUPPORTED_CLOCK_SOURCES.index(source))
        args.append(rate)
        args.append(reset)
        cls._execute_command(unit, 0, args)

    @classmethod
    def get_clock(cls, unit):
        params = cls._execute_command(unit, 1, None)
        if params[0] >= len(EftInfo.SUPPORTED_CLOCK_SOURCES):
            raise OSError('Unexpected clock source in response')
        if params[1] not in EftInfo.SUPPORTED_SAMPLING_RATES:
            raise OSError('Unexpected sampling rate in response')
        return (params[1], EftInfo.SUPPORTED_CLOCK_SOURCES[params[0]])

    @classmethod
    def set_box_states(cls, unit, states):
        mask_set = 0
        mask_clear = 0
        for name, state in states.items():
            if name not in cls.SUPPORTED_BOX_STATES:
                raise ValueError('Invalid value in box states')
            shift = cls.__BOX_STATE_POSITIONS[name]
            if cls.SUPPORTED_BOX_STATES[name].index(state) == 0:
                mask_clear |= (1 << shift)
            else:
                mask_set |= (1 << shift)
        args = array('I')
        args.append(mask_set)
        args.append(mask_clear)
        cls._execute_command(unit, 3, args)

    @classmethod
    def get_box_states(cls, unit):
        params = cls._execute_command(unit, 4, None)
        states = {}
        for name, shift in cls.__BOX_STATE_POSITIONS.items():
            index = (params[0] >> shift) & 0x01
            states[name] = cls.SUPPORTED_BOX_STATES[name][index]
        return states

    @classmethod
    def reconnect_phy(cls, unit):
        cls._execute_command(unit, 6, None)

    @classmethod
    def blink_leds(cls, unit):
        cls._execute_command(unit, 7, None)

    @classmethod
    def set_continuous_clock(cls, unit, continuous_rate):
        args = array('I')
        args.append(continuous_rate * 512 // 1500)
        cls._execute_command(unit, 8, args)

#
# Category No.4, for physical output multiplexer commands
#


class EftPhysOutput():
    OPERATIONS = ('gain', 'mute', 'nominal')

    @staticmethod
    def _execute_command(unit, cmd, args):
        if not isinstance(unit, Hinawa.SndEfw):
            raise ValueError('Invalid argument for SndEfw')
        params = [0] * 256
        return unit.transaction(4, cmd, args, params)

    @classmethod
    def set_param(cls, unit, operation, channel, value):
        if operation == 'gain':
            cmd = 0
        elif operation == 'mute':
            cmd = 2
            if value > 0:
                value = 1
        elif operation == 'nominal':
            cmd = 8
            if value > 0:
                value = 2
        else:
            raise ValueError('Invalid argument for operation.')
        args = array('I')
        args.append(channel)
        args.append(value)
        cls._execute_command(unit, cmd, args)

    @classmethod
    def get_param(cls, unit, operation, channel):
        if operation == 'gain':
            cmd = 1
        elif operation == 'mute':
            cmd = 3
        elif operation == 'nominal':
            print('Unfortunately, this doesn\'t work well...')
            cmd = 9
        else:
            raise ValueError('Invalid argument for operation.')
        args = array('I')
        args.append(channel)
        params = cls._execute_command(unit, cmd, args)
        if operation == 'nominal':
            if params[1] == 2:
                params[1] = 1
        return params[1]

#
# Category No.5, for physical input multiplexer commands
#


class EftPhysInput():
    OPERATIONS = ('nominal')

    @staticmethod
    def _execute_command(unit, cmd, args):
        if not isinstance(unit, Hinawa.SndEfw):
            raise ValueError('Invalid argument for SndEfw')
        params = [0] * 256
        return unit.transaction(5, cmd, args, params)

    @classmethod
    def set_param(cls, unit, operation, channel, value):
        if operation == 'nominal':
            cmd = 8
            if value > 0:
                value = 2
        else:
            raise ValueError('Invalid argument for operation')
        args = array('I')
        args.append(channel)
        args.append(value)
        cls._execute_command(unit, cmd, args)

    @classmethod
    def get_param(cls, unit, operation, channel):
        if operation == 'nominal':
            print('Unfortunately, this doesn\'t work well...')
            cmd = 9
        else:
            raise ValueError('Invalid argumentfor operation')
        args = array('I')
        args.append(channel)
        args.append(0xff)
        params = cls._execute_command(unit, cmd, args)
        return params[1]

#
# Category No.6, for playback stream multiplexer commands
#


class EftPlayback():
    OPERATIONS = ('gain', 'mute', 'solo')

    @staticmethod
    def _execute_command(unit, cmd, args):
        if not isinstance(unit, Hinawa.SndEfw):
            raise ValueError('Invalid argument for SndEfw')
        params = [0] * 256
        return unit.transaction(6, cmd, args, params)

    @classmethod
    def set_param(cls, unit, operation, channel, value):
        if operation == 'gain':
            cmd = 0
        elif operation == 'mute':
            cmd = 2
            if value > 0:
                value = 1
        elif operation == 'solo':
            cmd = 4
            if value > 0:
                value = 1
        else:
            raise ValueError('Invalid argument for operation.')
        args = array('I')
        args.append(channel)
        args.append(value)
        cls._execute_command(unit, cmd, args)

    @classmethod
    def get_param(cls, unit, operation, channel):
        if operation == 'gain':
            cmd = 1
        elif operation == 'mute':
            cmd = 3
        elif operation == 'solo':
            cmd = 5
        else:
            raise ValueError('Invalid argument for operation.')
        args = array('I')
        args.append(channel)
        params = cls._execute_command(unit, cmd, args)
        return params[1]


class EftCapture():
    OPERATIONS = ()

    @staticmethod
    def _execute_command(unit, cmd, args):
        if not isinstance(unit, Hinawa.SndEfw):
            raise ValueError('Invalid argument for SndEfw')
        params = [0] * 256
        return unit.transaction(7, cmd, args, params)

#
# Category No.8, for input monitoring multiplexer commands
#


class EftMonitor():
    OPERATIONS = ('gain', 'mute', 'solo', 'pan')

    @staticmethod
    def _execute_command(unit, cmd, args):
        if not isinstance(unit, Hinawa.SndEfw):
            raise ValueError('Invalid argument for SndEfw')
        params = [0] * 256
        return unit.transaction(8, cmd, args, params)

    @classmethod
    def set_param(cls, unit, operation, in_ch, out_ch, value):
        if operation == 'gain':
            cmd = 0
        elif operation == 'mute':
            cmd = 2
            if value > 0:
                value = 1
        elif operation == 'solo':
            cmd = 4
            if value > 0:
                value = 1
        elif operation == 'pan':
            cmd = 6
            if value < 0 or value > 255:
                raise ValueError('Invalid argument for panning')
        else:
            raise ValueError('Invalid argument for operation.')
        args = array('I')
        args.append(in_ch)
        args.append(out_ch)
        args.append(value)
        cls._execute_command(unit, cmd, args)

    @classmethod
    def get_param(cls, unit, operation, in_ch, out_ch):
        if operation == 'gain':
            cmd = 1
        elif operation == 'mute':
            cmd = 3
        elif operation == 'solo':
            cmd = 5
        elif operation == 'pan':
            cmd = 7
        else:
            raise ValueError('Invalid argument for operation.')
        args = array('I')
        args.append(in_ch)
        args.append(out_ch)
        params = cls._execute_command(unit, cmd, args)
        return params[2]

#
# Category No.9, for input/output configuration commands
#


class EftIoconf():
    # NOTE: use the same strings in features of EftInfo.
    DIGITAL_INPUT_MODES = ('spdif-coax', 'aesebu-xlr', 'spdif-opt', 'adat-opt')

    @staticmethod
    def _execute_command(unit, cmd, args):
        if not isinstance(unit, Hinawa.SndEfw):
            raise ValueError('Invalid argument for SndEfw')
        params = [0] * 256
        return unit.transaction(9, cmd, args, params)

    @classmethod
    def set_control_room_mirroring(cls, unit, output_pair):
        args = array('I')
        args.append(output_pair)
        cls._execute_command(unit, 0, args)

    @classmethod
    def get_control_room_mirroring(cls, unit):
        params = cls._execute_command(unit, 1, None)
        return params[0]

    @classmethod
    def set_digital_input_mode(cls, unit, mode):
        if mode not in cls.DIGITAL_INPUT_MODES:
            raise ValueError('Invalid argument for digital mode')
        args = array('I')
        args.append(cls.DIGITAL_INPUT_MODES.index(mode))
        cls._execute_command(unit, 2, args)

    @classmethod
    def get_digital_input_mode(cls, unit):
        params = cls._execute_command(unit, 3, None)
        if params[0] >= len(cls.DIGITAL_INPUT_MODES):
            raise OSError
        return cls.DIGITAL_INPUT_MODES[params[0]]

    @classmethod
    def set_phantom_powering(cls, unit, state):
        if state > 0:
            state = 1
        args = array('I')
        args.append(state)
        cls._execute_command(unit, 4, args)

    @classmethod
    def get_phantom_powering(cls, unit):
        params = cls._execute_command(unit, 5, None)
        return params[0]

    @classmethod
    def set_stream_mapping(cls, unit, rx_maps, tx_maps):
        params = cls._execute_command(unit, 7, None)
        rx_map_count = params[2]
        if len(rx_maps) > rx_map_count:
            ValueError('Invalid argument for rx stream mapping')
        tx_map_count = params[34]
        if len(tx_maps) > tx_map_count:
            ValueError('Invalid argument for tx stream mapping')
        for i in range(len(rx_maps)):
            params[4 + i] = rx_maps[i]
        for i in range(len(tx_maps)):
            params[36 + i] = tx_maps[i]
        cls._execute_command(unit, 6, params)

    @classmethod
    def get_stream_mapping(cls, unit):
        param = cls._execute_command(unit, 7, None)
        tx_map_count = param[34]
        tx_map = []
        for i in range(tx_map_count):
            tx_map.append(param[36 + i])
        rx_map_count = param[2]
        rx_map = []
        for i in range(rx_map_count):
            rx_map.append(param[4 + i])
        return {'tx-map': tx_map, 'rx-map': rx_map}



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.bebob.maudio_protocol_normal import MaudioProtocolNormal

from hinawa_utils.ta1394.audio import AvcAudio

__all__ = ['MaudioProtocolNormal']


class MaudioProtocolFw410(MaudioProtocolNormal):
    __SELECTOR_FB = 0x07
    __PROCESSING_OUT_FB = 0x07
    __PROCESSING_IN_FB = 0x00
    __PROCESSING_OUT_CH = 0x01

    def set_headphone_source(self, target, source):
        if target not in self.get_headphone_labels():
            raise ValueError('Invalid argument for headphone')
        if source not in self.get_headphone_source_labels(target):
            raise ValueError('Invalid argument for headphone source')

        if source == 'aux-1/2':
            value = 0x01
        else:
            value = 0x00
        AvcAudio.set_selector_state(self.unit.fcp, 0, 'current',
                                    self.__SELECTOR_FB, value)

        if source != 'aux-1/2':
            for i, elems in enumerate(self.mixers):
                if self.labels['mixers'][i] == source:
                    data = (0x00, 0x00)
                else:
                    data = (0x80, 0x00)

                in_ch = elems[1][0]
                AvcAudio.set_processing_mixer_state(self.unit.fcp, 0,
                                                    'current', self.__PROCESSING_OUT_FB,
                                                    self.__PROCESSING_IN_FB, in_ch,
                                                    self.__PROCESSING_OUT_CH, data)

    def get_headphone_source(self, target):
        if target not in self.get_headphone_labels():
            raise ValueError('Invalid argument for headphone')

        state = AvcAudio.get_selector_state(self.unit.fcp, 0, 'current',
                                            self.__SELECTOR_FB)
        if state == 0x01:
            return 'aux-1/2'
        elif state == 0x00:
            for i, elems in enumerate(self.mixers):
                in_ch = elems[1][0]
                data = AvcAudio.get_processing_mixer_state(self.unit.fcp, 0,
                                                           'current',
                                                           self.__PROCESSING_OUT_FB, self.__PROCESSING_IN_FB,
                                                           in_ch, self.__PROCESSING_OUT_CH)
                if data[0] == 0x00 and data[1] == 0x00:
                    return self.labels['mixers'][i]

        # No sources are set. For convenience, set 'mixer-1/2'.
        self.set_headphone_source('headphone-1/2', 'mixer-1/2')
        return 'mixer-1/2'



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from struct import unpack

from hinawa_utils.ieee1212.config_rom_lexer import Ieee1212ConfigRomLexer
from hinawa_utils.ieee1212.root_directory_parser import Ieee1212RootDirectoryParser

__all__ = ['Ieee1394ConfigRomParser']


class Ieee1394ConfigRomParser(Ieee1212RootDirectoryParser):
    _NAME = '1394'

    __BUS_CAPABILITIES_1995 = {
        'imc':  7,  # the node is IRM capable.
        'cmc':  6,  # the node is cycle master capable.
        'isc':  5,  # the node supports isochronous operations.
        'bmc':  4,  # the node is bus namager capable.
    }

    # added by IEEE 1394:2008
    __BUS_CAPABILITIES_2008 = {
        'pmc':  3,  # the node is power manager capable.
        'adj':  2,  # the node is compliant to IEEE 1394.1:2004.
    }

    # IEEE 1394:1995 refers to ISO/IEC 13213:1994 (ANSI/IEEE Std 1212:1994).
    __NODE_CAPABILITIES = {
        'misc': {
            'spt':  15,  # The SPLIT_TIMEOUT register is implemented.
            'ms':  14,  # The messages-passing registers are implemented.
            'int':  13,  # The INTERRUPT_TARGET and INTERRUPT_MASK registers are
                        # implemented.
        },
        'testing': {
            'ext':  12,  # The ARGUMENT registers are implemented.
            'bas':  11,  # Node implements TEST_START&TEST_STATUS registers and
            # testing state.
        },
        'addressing': {
            'prv':  10,  # The node implements the private space.
            '64':   9,  # The node uses 64-bit aaddressing (otherwise 32-bit
                        # addressing).
            'fix':  8,  # The node uses the fixed addressing scheme (otherwise
                        # extended addressing).
        },
        'state': {
            'lst':  7,  # The STATE_BITS.lost bit is implemented.
            'drq':  6,  # The STATE_BITS.dreq bit is implemented.
            'elo':  4,  # The STATE_BITS.elog bit and the ERROR_LOG registers
                        # are implementd.
            'atn':  3,  # The STATE_BITS.atn bit is implemented.
            'off':  2,  # The STATE_BITS.off bit is implemented.
            'ded':  1,  # The node supports the dead state.
            'init': 0,  # The node supports the initializing state.
        },
    }

    def __init__(self):
        super().__init__()

    def _parse_ieee1394_bus_info(self, data):
        info = {}

        name = data[0:4].decode('US-ASCII')
        if name != self._NAME:
            raise ValueError(
                'Invalid data for Configuration ROM in IEEE 1394.')

        info['name'] = name

        for name, shift in self.__BUS_CAPABILITIES_1995.items():
            info[name] = bool(data[4] & (1 << shift))

        info['cyc_clk_acc'] = data[5]

        val = data[6] >> 4
        info['max_rec'] = 0 if val == 0 else pow(2, val + 1)

        # reserved fields in IEEE 1394:1995
        if data[4] & 0x0f or data[6] & 0x04 or data[7]:
            # IEEE 1394:2008
            for name, shift in self.__BUS_CAPABILITIES_2008.items():
                info[name] = bool(data[4] & (1 << shift))
            info['max_ROM'] = data[6] & 0x03
            info['generation'] = data[7] >> 4
            info['link_spd'] = data[7] & 0x07

        info['node_vendor_ID'] = (unpack('>H', data[8:10])[0] << 8) | data[10]
        info['chip_ID'] = (data[11] << 32) | unpack('>I', data[12:16])[0]

        return info

    def _handle_bus_dep_keys(self, key_name, type_name, data):
        if key_name == 'NODE_CAPABILITIES' and type_name == 'IMMEDIATE':
            info = {}
            for kind, elems in self.__NODE_CAPABILITIES.items():
                if kind not in info:
                    info[kind] = {}
                for name, shift in elems.items():
                    info[kind][name] = bool(data & (1 << shift))
            return info
        return None

    def parse_rom(self, data):
        info = {}

        entries = Ieee1212ConfigRomLexer.detect_entries(data)

        bus_info = entries['bus-info']
        info['bus-info'] = self._parse_ieee1394_bus_info(bus_info)

        self.add_bus_dep_handle(self._NAME, self._handle_bus_dep_keys)
        root = entries['root-directory']
        info['root-directory'] = self.parse_root_directory(self._NAME, root)

        return info



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.dice.tcat_protocol_extension import ExtMixerSpace, ExtCurrentConfigSpace


class TcatTcd22xxSpec():
    # These four properties should be overridden by derived classes.
    MODELS = (
        (0x000000, 0x000000),
    )
    _INPUTS = (
        (),
    )
    _OUTPUTS = (
        (),
    )
    _FIXED = (
        {},
    )

    def __init__(self, index):
        self._index = index

    def _get_available_mixer_ports(self, protocol, mode):
        dsts = []

        count = 0
        for id, max_ports in ExtMixerSpace.MIXER_IN_MAX_PORTS.items():
            if count + max_ports > protocol._ext_caps['mixer']['input-channels']:
                max_ports = protocol._ext_caps['mixer']['input-channels'] - count
            for i in range(0, max_ports, 2):
                label = 'Mixer-{0}/{1}'.format(count + i + 1, count + i + 2)
                dst = [label, id, [i, i + 1]]
                dsts.append(dst)
            count += max_ports

        srcs = []

        max_ports = min(ExtMixerSpace.MIXER_OUT_MAX_PORTS[mode],
                        protocol._ext_caps['mixer']['output-channels'])
        for i in range(0, max_ports, 2):
            label = 'Mixer-{0}/{1}'.format(i + 1, i + 2)
            src = [label, 'mixer', [i, i + 1]]
            srcs.append(src)

        return srcs, dsts

    def _get_available_stream_ports(self, protocol, req, mode):
        STREAMS = ('avs0', 'avs1')

        stream_configs = \
            ExtCurrentConfigSpace.read_stream_config(protocol, req, mode)

        dsts = []

        for i, params in enumerate(stream_configs['tx']):
            id = STREAMS[i]
            count = params['pcm']
            if len(stream_configs['tx']) == 1:
                suffix = ''
            else:
                suffix = '-{0:c}'.format(0x41 + i)
            for j in range(0, count, 2):
                label = 'Stream{0}-{1}/{2}'.format(suffix, j + 1, j + 2)
                dst = [label, id, [j, j + 1]]
                dsts.append(dst)

        srcs = []

        for i, params in enumerate(stream_configs['rx']):
            id = STREAMS[i]
            count = params['pcm']
            if len(stream_configs['rx']) == 1:
                suffix = ''
            else:
                suffix = '-{0:c}'.format(0x41 + i)
            for j in range(0, count, 2):
                label = 'Stream{0}-{1}/{2}'.format(suffix, j + 1, j + 2)
                src = [label, id, [j, j + 1]]
                srcs.append(src)

        return srcs, dsts

    def _get_available_virt_ports(self, protocol, req, mode):
        srcs = []
        dsts = []

        stream_srcs, stream_dsts = \
            self._get_available_stream_ports(protocol, req, mode)
        srcs.extend(stream_srcs)
        dsts.extend(stream_dsts)

        mixer_srcs, mixer_dsts = self._get_available_mixer_ports(
            protocol, mode)
        srcs.extend(mixer_srcs)
        dsts.extend(mixer_dsts)

        return srcs, dsts

    def _get_full_spec_ports(self, protocol):
        inputs = []
        outputs = []

        asic_type = protocol._ext_caps['general']['asic-type']
        if asic_type == 'TCD-2210':
            suffix = ''
        else:
            suffix = '-A'

        inputs.append(('Analog{0}'.format(suffix), 'ins0', 16))
        outputs.append(('Analog{0}'.format(suffix), 'ins0', 16))

        if suffix != '':
            inputs.append(('Analog-B', 'ins1', 8))
            outputs.append(('Analog-B', 'ins1', 8))

        adat_count = 0
        for clock in ('adat', 'tdif'):
            if clock in protocol._clock_sources:
                adat_count += 1

        if adat_count == 1:
            inputs.append(('ADAT', 'adat', 8))
            outputs.append(('ADAT', 'adat', 8))
        elif adat_count == 2 and asic_type == 'TCD-2220':
            inputs.append(('ADAT-A', 'adat', 8))
            inputs.append(('ADAT-B', 'adat', 8))
            outputs.append(('ADAT-A', 'adat', 8))
            outputs.append(('ADAT-B', 'adat', 8))

        count = 0
        has_suffix = False
        for src in protocol._clock_sources:
            if src.find('aes') == 0:
                count += 1
                if count > 1:
                    has_suffix = True
        count = 0
        for src in protocol._clock_sources:
            if src.find('aes') == 0:
                index = {v: k for k, v in protocol.CLOCK_BITS.items()}[src]
                name = protocol._clock_source_labels[index]
                if has_suffix:
                    label = '{0}-{1:c}'.format(name, count + 0x41)
                else:
                    label = name
                inputs.append((label, 'aes', 2))
                outputs.append((label, 'aes', 2))
                count += 1

        return inputs, outputs

    def _get_available_real_ports(self, protocol, mode, inputs, outputs):
        ADAT_CHS = {
            'low': 8,
            'middle': 4,
            'high': 2,
        }

        # Update cache of physical ports. ADAT ports reduce the number of
        # available channels.
        in_ports = []
        out_ports = []
        for i, entry in enumerate(inputs):
            params = list(entry)
            if params[1] == 'adat':
                params[2] = ADAT_CHS[mode]
            in_ports.append(params)
        for i, entry in enumerate(outputs):
            params = list(entry)
            if params[1] == 'adat':
                params[2] = ADAT_CHS[mode]
            out_ports.append(params)

        # Some models have offset in their aes entries.
        aes_offset = 0
        clocks = protocol.get_supported_clock_sources()
        for clock in clocks:
            if clock.find('aes') == 0:
                for key, val in protocol.CLOCK_BITS.items():
                    if val == clock:
                        aes_offset = key * 2
                        break
                break

        dsts = []

        for params in out_ports:
            id = params[1]
            count = params[2]
            if id.find('aes') != 0:
                offset = 0
            else:
                offset = aes_offset
            for dst in dsts:
                if id == dst[1]:
                    offset += 2
            for i in range(0, count, 2):
                label = '{0}-{1}/{2}'.format(params[0], i + 1, i + 2)
                dst = [label, id, [offset + i, offset + i + 1]]
                dsts.append(dst)

        srcs = []

        for params in in_ports:
            id = params[1]
            count = params[2]
            if id.find('aes') != 0:
                offset = 0
            else:
                offset = aes_offset
            for src in srcs:
                if id == src[1]:
                    offset += 2
            for i in range(0, count, 2):
                label = '{0}-{1}/{2}'.format(params[0], i + 1, i + 2)
                src = [label, id, [offset + i, offset + i + 1]]
                srcs.append(src)

        return srcs, dsts

    def get_available_ports(self, protocol, req, mode):
        srcs = []
        dsts = []

        inputs = self._INPUTS[self._index]
        outputs = self._OUTPUTS[self._index]

        if len(inputs) == 0 and len(outputs) == 0:
            inputs, outputs = self._get_full_spec_ports(protocol)

        # Check arguments.
        for port in inputs:
            if len(port) != 3:
                raise ValueError('Invalid entry for input port.')
        for port in outputs:
            if len(port) != 3:
                raise ValueError('Invalid entry for output port.')

        real_srcs, real_dsts = \
            self._get_available_real_ports(protocol, mode, inputs,
                                           outputs)
        srcs.extend(real_srcs)
        dsts.extend(real_dsts)

        virt_srcs, virt_dsts = \
            self._get_available_virt_ports(protocol, req, mode)
        srcs.extend(virt_srcs)
        dsts.extend(virt_dsts)

        return srcs, dsts

    def _refine_entries(self, entries, srcs, dsts):
        routes = []

        for entry in entries:
            # Skip mixer-to-mixer entries.
            if (entry['src-blk'] == 'mixer' and
                    entry['dst-blk'] in ('mixer-tx0', 'mixer-tx1')):
                continue

            # Skip entries with mute or reserved.
            if entry['dst-blk'] == 'mute':
                continue
            if entry['src-blk'].find('reserved') == 0:
                continue

            # Skip invalid entries for source.
            for src in srcs:
                if entry['src-blk'] == src[1] and entry['src-ch'] in src[2]:
                    break
            else:
                continue

            # Skip invalid entries for destination.
            for dst in dsts:
                if entry['dst-blk'] == dst[1] and entry['dst-ch'] in dst[2]:
                    break
            else:
                continue

            for i in range(2):
                route = {
                    'src-blk':  src[1],
                    'src-ch':   src[2][i],
                    'dst-blk':  dst[1],
                    'dst-ch':   dst[2][i],
                    'peak':     0,
                }
                if route not in routes:
                    routes.append(route)

        return routes

    def _add_fixed_entries(self, routes, maximum):
        for index, params in self._FIXED[self._index].items():
            direction = params[0]
            if direction == 'src':
                entry = {
                    'src-blk':  params[1],
                    'src-ch':   params[2],
                    'dst-blk':  'reserved6',
                    'dst-ch':   params[2] % 2,
                    'peak':     0,
                }
            elif direction == 'dst':
                entry = {
                    'src-blk':  'mute',
                    'src-ch':   params[2] % 2,
                    'dst-blk':  params[1],
                    'dst-ch':   params[2],
                    'peak':     0,
                }

            if (index < len(routes) and routes[index] != entry and
                    len(routes) < maximum):
                routes.insert(index, entry)

        return routes

    def normalize_router_entries(self, protocol, entries, srcs, dsts):
        maximum = protocol._ext_caps['router']['maximum-routes']

        # Reset peak meter.
        for entry in entries:
            entry['peak'] = 0

        # Refine with missing pairs.
        routes = self._refine_entries(entries, srcs, dsts)

        # Some models require entries in fixed positions to display level in
        # its physical metrs.
        routes = self._add_fixed_entries(routes, maximum)

        return routes



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.bebob.bebob_unit import BebobUnit

from hinawa_utils.bebob.phase_go_protocol_coax import PhaseGoProtocolCoax
from hinawa_utils.bebob.phase_go_protocol_opt import PhaseGoProtocolOpt

__all__ = ['PhaseGoUnit']


class PhaseGoUnit(BebobUnit):
    _SUPPORTED_MODELS = {
        # (VendorID, ModelID): ProtocolClass
        (0x000aac, 0x000004):   PhaseGoProtocolCoax,    # Terratec PHASE 24 FW
        (0x000aac, 0x000007):   PhaseGoProtocolOpt,     # Terratec PHASE X24 FW
        (0x00a0de, 0x10000b):   PhaseGoProtocolCoax,    # Yamaha Go44
        (0x00a0de, 0x10000c):   PhaseGoProtocolOpt,     # Yamaha Go46
    }

    def __init__(self, path):
        super().__init__(path)

        key = (self.vendor_id, self.model_id)
        if key not in self._SUPPORTED_MODELS:
            raise OSError('Not supported.')
        self.protocol = self._SUPPORTED_MODELS[key](self.fcp)



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.ta1394.general import AvcGeneral
from hinawa_utils.ta1394.general import AvcConnection
from hinawa_utils.ta1394.ccm import AvcCcm

from hinawa_utils.bebob.bebob_unit import BebobUnit
from hinawa_utils.bebob.extensions import BcoPlugInfo
from hinawa_utils.bebob.extensions import BcoSubunitInfo
from hinawa_utils.bebob.extensions import BcoStreamFormatInfo

__all__ = ['PlugParser']


class PlugParser():
    @classmethod
    def parse_unit_info(cls, fcp):
        return AvcGeneral.get_unit_info(fcp)

    @classmethod
    def parse_unit_plugs(cls, fcp):
        unit_plugs = {}
        info = AvcConnection.get_unit_plug_info(fcp)
        for type, params in info.items():
            if type not in unit_plugs:
                unit_plugs[type] = {}
                unit_plugs[type]['output'] = {}
                unit_plugs[type]['input'] = {}
            for dir, num in params.items():
                for i in range(num + 1):
                    try:
                        plug = cls.parse_unit_plug(fcp, dir, type, i)
                        unit_plugs[type][dir][i] = plug
                    except Exception:
                        continue
        return unit_plugs

    @classmethod
    def parse_unit_plug(cls, fcp, dir, type, num):
        plug = {}
        addr = BcoPlugInfo.get_unit_addr(dir, type, num)
        plug['type'] = BcoPlugInfo.get_plug_type(fcp, addr)
        plug['name'] = BcoPlugInfo.get_plug_name(fcp, addr)
        plug['channels'] = []
        channels = BcoPlugInfo.get_plug_channels(fcp, addr)
        for channel in range(channels):
            ch = BcoPlugInfo.get_plug_ch_name(fcp, addr, channel + 1)
            plug['channels'].append(ch)
        plug['clusters'] = []
        if plug['type'] == 'IsoStream':
            clusters = BcoPlugInfo.get_plug_clusters(fcp, addr)
            for cluster in range(len(clusters)):
                clst = BcoPlugInfo.get_plug_cluster_info(
                    fcp, addr, cluster + 1)
                plug['clusters'].append(clst)
        plug['input'] = []
        plug['outputs'] = []
        if dir == 'output':
            plug['input'] = BcoPlugInfo.get_plug_input(fcp, addr)
        else:
            plug['outputs'] = BcoPlugInfo.get_plug_outputs(fcp, addr)
        return plug

    @classmethod
    def parse_subunit_plugs(cls, fcp):
        subunit_plugs = {}
        for page in range(AvcGeneral.MAXIMUM_SUBUNIT_PAGE + 1):
            try:
                subunits = AvcGeneral.get_subunit_info(fcp, page)
            except Exception:
                break

            for entry in subunits:
                type = entry['type']
                maximum_id = entry['maximum-id']
                if type not in subunit_plugs:
                    subunit_plugs[type] = {}
                for id in range(maximum_id + 1):
                    if id not in subunit_plugs[type]:
                        subunit_plugs[type][id] = {}
                        subunit_plugs[type][id]['output'] = {}
                        subunit_plugs[type][id]['input'] = {}

                info = AvcConnection.get_subunit_plug_info(fcp, type, 0)
                for dir, num in info.items():
                    for i in range(num):
                        plug = cls.parse_subunit_plug(fcp, dir, type, 0, i)
                        subunit_plugs[type][id][dir][i] = plug
        return subunit_plugs

    @classmethod
    def parse_subunit_plug(cls, fcp, dir, type, id, num):
        plug = {}
        addr = BcoPlugInfo.get_subunit_addr(dir, type, id, num)
        plug['type'] = BcoPlugInfo.get_plug_type(fcp, addr)
        plug['name'] = BcoPlugInfo.get_plug_name(fcp, addr)
        plug['channels'] = []
        channels = BcoPlugInfo.get_plug_channels(fcp, addr)
        for channel in range(channels):
            ch = BcoPlugInfo.get_plug_ch_name(fcp, addr, channel + 1)
            plug['channels'].append(ch)
        plug['clusters'] = []
        if plug['type'] == 'IsoStream':
            clusters = BcoPlugInfo.get_plug_clusters(fcp, addr)
            for cluster in range(len(clusters)):
                clst = BcoPlugInfo.get_plug_cluster_info(
                    fcp, addr, cluster + 1)
                plug['clusters'].append(clst)
        plug['input'] = {}
        plug['outputs'] = []
        # Music subunits have counter direction.
        try:
            plug['input'] = BcoPlugInfo.get_plug_input(fcp, addr)
        except Exception:
            pass
        try:
            plug['outputs'] = BcoPlugInfo.get_plug_outputs(fcp, addr)
        except Exception:
            pass
        return plug

    @classmethod
    def parse_function_block_plugs(cls, fcp, subunit_plugs):
        subunits = {}
        for subunit_type, subunit_type_plugs in subunit_plugs.items():
            if subunit_type not in subunits:
                subunits[subunit_type] = {}

            for subunit_id in subunit_type_plugs.keys():
                fbs = {}

                entries = []
                for page in range(0xff):
                    elems = BcoSubunitInfo.get_subunit_fb_info(fcp,
                                                               subunit_type, subunit_id, page, 0xff)
                    if len(elems) == 0:
                        break
                    entries.extend(elems)

                for entry in entries:
                    fb_type = entry['type']
                    if fb_type not in fbs:
                        fbs[fb_type] = {}
                    fb_id = entry['id']

                    fb = {}
                    fb['purpose'] = entry['purpose']
                    fb['outputs'] = {}
                    fb['inputs'] = {}
                    for i in range(entry['inputs']):
                        plug = cls.parse_fb_plug(fcp, 'input', subunit_type, subunit_id, fb_type, fb_id, i)
                        fb['inputs'][i] = plug
                    for i in range(entry['outputs']):
                        plug = cls.parse_fb_plug(fcp, 'output', subunit_type, subunit_id, fb_type, fb_id, i)
                        fb['outputs'][i] = plug

                    fbs[fb_type][fb_id] = fb

                subunits[subunit_type][subunit_id] = fbs

        return subunits

    @classmethod
    def parse_fb_plug(cls, fcp, dir, subunit_type, subunit_id, fb_type, fb_id, num):
        plug = {}
        addr = BcoPlugInfo.get_function_block_addr(dir, subunit_type,
                                                   subunit_id, fb_type, fb_id, num)
        plug['type'] = BcoPlugInfo.get_plug_type(fcp, addr)
        plug['name'] = BcoPlugInfo.get_plug_name(fcp, addr)
        plug['channels'] = []
        channels = BcoPlugInfo.get_plug_channels(fcp, addr)
        for channel in range(channels):
            ch = BcoPlugInfo.get_plug_ch_name(fcp, addr, channel + 1)
            plug['channels'].append(ch)
        plug['clusters'] = []
        if plug['type'] == 'IsoStream':
            clusters = BcoPlugInfo.get_plug_clusters(fcp, addr)
            for cluster in range(len(clusters)):
                clst = BcoPlugInfo.get_plug_cluster_info(
                    fcp, addr, cluster + 1)
                plug['clusters'].append(clst)
        plug['input'] = {}
        plug['outputs'] = []
        # Music subunits have counter direction.
        try:
            plug['input'] = BcoPlugInfo.get_plug_input(fcp, addr)
        except Exception:
            pass
        try:
            plug['outputs'] = BcoPlugInfo.get_plug_outputs(fcp, addr)
        except Exception:
            pass
        return plug

    @classmethod
    def parse_signal_destination(cls, fcp, subunit_plugs):
        dst = []
        for subunit_id, subunit_id_plugs in subunit_plugs['music'].items():
            for i, plug in subunit_id_plugs['input'].items():
                if plug['type'] == 'Sync':
                    dst = AvcCcm.get_subunit_signal_addr('music', 0, i)
        return dst

    @classmethod
    def parse_signal_sources(cls, fcp, unit_plugs, subunit_plugs, signal_destination):
        srcs = []
        candidates = []
        # This is internal clock source.
        for subunit_id, subunit_id_plugs in subunit_plugs['music'].items():
            for i, plug in subunit_id_plugs['output'].items():
                if plug['type'] == 'Sync':
                    addr = AvcCcm.get_subunit_signal_addr('music', 0, i)
                    candidates.append((addr, plug))
        # External source is available.
        for i, plug in unit_plugs['external']['input'].items():
            if plug['type'] in ('Sync', 'Digital', 'Clock'):
                addr = AvcCcm.get_unit_signal_addr('external', i)
                candidates.append((addr, plug))
        # SYT-match is available, but not practical.
        for i, plug in unit_plugs['isoc']['input'].items():
            if plug['type'] == 'Sync':
                addr = AvcCcm.get_unit_signal_addr('isoc', i)
                candidates.append((addr, plug))
        # BeBoBv3 has.
        # Inquire these are able to connect to destination.
        for params in candidates:
            addr = params[0]
            plug = params[1]
            try:
                AvcCcm.ask_signal_source(fcp, addr, signal_destination)
            except Exception:
                continue
            srcs.append(params)
        return srcs

    @classmethod
    def parse_stream_formats(cls, fcp, unit_plugs):
        formats = {}
        for type, dir_plugs in unit_plugs.items():
            if type == 'async':
                continue
            formats[type] = {}
            for dir, plugs in dir_plugs.items():
                formats[type][dir] = {}
                for i, plug in plugs.items():
                    addr = BcoPlugInfo.get_unit_addr(dir, type, i)
                    try:
                        fmts = BcoStreamFormatInfo.get_entry_list(fcp,
                                                                  addr)
                        formats[type][dir][i] = fmts
                    except Exception:
                        continue
        return formats

    @classmethod
    def get_unit_plug_list(cls, fcp):
        plugs = {}
        seqid = 0

        units = AvcConnection.get_unit_plug_info(fcp)
        for type, data in units.items():
            for direction, count in data.items():
                for plug_id in range(count):
                    # Use the same format as Plug Input/Output Specific Data
                    # to keep enough informaton.
                    plug_info = {
                        'dir': direction,
                        'mode': 'unit',
                        'data': {
                            'unit-type': type,
                            'plug': plug_id,
                        },
                    }
                    plugs['unit_{0}'.format(seqid)] = plug_info
                    seqid += 1

        return plugs

    @classmethod
    def get_subunit_plug_info(cls, fcp):
        subunits = {}
        for page in range(AvcGeneral.MAXIMUM_SUBUNIT_PAGE + 1):
            try:
                info = AvcGeneral.get_subunit_info(fcp, page)
            except Exception:
                break

            for entry in info:
                subunit_type = entry['type']
                maximum_id = entry['maximum-id']

                for subunit_id in range(maximum_id + 1):
                    try:
                        data = AvcConnection.get_subunit_plug_info(fcp,
                                                    subunit_type, subunit_id)
                    except Exception:
                        continue

                    id = (subunit_type, subunit_id)
                    if id not in subunits:
                        subunits[id] = {}
                    for direction, count in data.items():
                        subunits[id][direction] = count

        return subunits

    @classmethod
    def get_subunit_plug_list(cls, fcp):
        plugs = {}
        seqid = 0

        subunits = cls.get_subunit_plug_info(fcp)
        for id, data in subunits.items():
            for direction, count in data.items():
                for plug_id in range(count):
                    # Use the same format as Plug Input/Output Specific Data
                    # to keep enough informaton.
                    plug_info = {
                        'dir': direction,
                        'mode': 'subunit',
                        'data': {
                            'subunit-type': id[0],
                            'subunit-id': id[1],
                            'plug': plug_id,
                        },
                    }
                    plugs['subunit_{0}'.format(seqid)] = plug_info
                    seqid += 1

        return plugs

    @classmethod
    def get_avail_connections(cls, fcp, unit_plug_list, subunit_plug_list):
        src_candidates = {}
        dst_candidates = {}
        avail = {}

        for seqid, info in unit_plug_list.items():
            data = info['data']
            addr = AvcCcm.get_unit_signal_addr(data['unit-type'], data['plug'])
            if info['dir'] == 'output':
                target = dst_candidates
            else:
                target = src_candidates
            target[seqid] = addr

        for seqid, info in subunit_plug_list.items():
            data = info['data']
            addr = AvcCcm.get_subunit_signal_addr(data['subunit-type'],
                                            data['subunit-id'], data['plug'])
            # Inverse direction against plugs of unit.
            if info['dir'] == 'output':
                target = src_candidates
            else:
                target = dst_candidates
            target[seqid] = addr

        for dst_seqid, dst_addr in dst_candidates.items():
            try:
                curr_src_info = AvcCcm.get_signal_source(fcp, dst_addr)
            except Exception:
                curr_src_info = None

            for src_seqid, src_addr in src_candidates.items():
                try:
                    AvcCcm.ask_signal_source(fcp, src_addr, dst_addr)
                except Exception:
                    continue

                if dst_seqid not in avail:
                    avail[dst_seqid] = []

                src_info = AvcCcm.parse_signal_addr(src_addr)
                avail[dst_seqid].append((src_seqid, src_info == curr_src_info))

        return avail

    @classmethod
    def get_plug_spec(cls, fcp, info):
        data = info['data']
        if info['mode'] == 'unit':
            addr = BcoPlugInfo.get_unit_addr(info['dir'],
                                             data['unit-type'], data['plug'])
        elif info['mode'] == 'subunit':
            addr = BcoPlugInfo.get_subunit_addr(info['dir'],
                    data['subunit-type'], data['subunit-id'], data['plug'])
        else:
            raise ValueError('Invalid mode of plug info.')

        spec = {
            'name': BcoPlugInfo.get_plug_name(fcp, addr),
            'type': BcoPlugInfo.get_plug_type(fcp, addr),
        }

        if info['dir'] == 'input':
            spec['input'] = BcoPlugInfo.get_plug_input(fcp, addr),
        else:
            spec['outputs'] = BcoPlugInfo.get_plug_outputs(fcp, addr),

        return spec

        if info['mode'] == 'unit':
            spec['clusters'] = []
            clusters = BcoPlugInfo.get_plug_clusters(fcp, addr)

            for i, cluster in enumerate(clusters):
                mapping = []
                name = BcoPlugInfo.get_plug_cluster_info(fcp, addr, i + 1)
                for info in cluster:
                    idx, pos = info
                    ch_name = BcoPlugInfo.get_plug_ch_name(fcp, addr, idx)
                    mapping.append(ch_name)
                entry = {
                    'name': name,
                    'channels': mapping,
                }
                spec['clusters'].append(entry)

        return spec



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.ieee1394.config_rom_parser import Ieee1394ConfigRomParser

__all__ = ['EfwConfigRomParser']


class EfwConfigRomParser(Ieee1394ConfigRomParser):
    __OUI_ECHO = 0x001486
    __MANUFACTURER_OUI = {
        0x000ff2:   'Loud Technologies Inc.',
        __OUI_ECHO:  'Echo Digital Audio Corporation',
    }

    def __init__(self):
        super().__init__()
        self.add_vendor_dep_handle(
            self.__OUI_ECHO, self.__handle_echoaudio_keys)

    def __handle_echoaudio_keys(self, key_id, type_name, data):
        if key_id != 0x08 or type_name != 'IMMEDIATE':
            return None
        if data in self.__MANUFACTURER_OUI:
            name = self.__MANUFACTURER_OUI[data]
        else:
            name = data
        return ['MANUFACTURER', name]

    def parse_rom(self, data):
        entries = super().parse_rom(data)
        return self.__parse_entries(entries['root-directory'])

    def __parse_entries(self, entries):
        # Typical layout.
        FIELDS = (
            ('VENDOR',              'vendor-id'),
            ('DESCRIPTOR',          'vendor-name'),
            ('MODEL',               'model-id'),
            ('DESCRIPTOR',          'model-name'),
            ('NODE_CAPABILITIES',   'node-capabilities'),
            ('EUI_64',              'guid'),
            ('UNIT',                None),
            ('MANUFACTURER',        'manufacturer'),
        )
        info = {}

        for i, field in enumerate(FIELDS):
            entry = entries[i]
            name, alt = field
            if entry[0] != name:
                raise OSError('Invalid format of config ROM.')
            if name == 'UNIT':
                # Check unit.
                entry = entry[1]
                if (entry[0] != ['SPECIFIER_ID', 0x00a02d] or
                    entry[1] != ['VERSION',      0x010000] or
                    entry[2] != ['MODEL',        info['model-id']] or
                        entry[3] != ['DESCRIPTOR',   info['model-name']]):
                    raise ValueError('Invalid data of config ROM.')
            else:
                info[alt] = entry[1]

        return info



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.dice.tcat_tcd22xx_spec import TcatTcd22xxSpec

__all__ = ['PresonusFirestudioSpec']


class PresonusFirestudioSpec(TcatTcd22xxSpec):
    MODELS = (
        (0x000a92, 0x000008),   # Firestudio 26x26
        (0x000a92, 0x00000b),   # Firestudio project
        (0x000a92, 0x00000c),   # Firestudio tube
        (0x000a92, 0x000011),   # Firestudio mobile
    )

    _INPUTS = (
        (),
        (),
        (),
        (
            ('Analog',  'ins0', 8),
            ('S/PDIF',  'aes',  2),
        ),
    )

    _OUTPUTS = (
        (),
        (),
        (),
        (
            ('Analog',  'ins0', 4),
            ('S/PDIF',  'aes',  2),
        ),
    )

    _FIXED = (
        {},
        {},
        {},
        {
            0: ('src', 'ins0', 0),
            1: ('src', 'ins0', 1),
        },
    )



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from struct import pack, unpack

import gi
gi.require_version('Hinawa', '3.0')
from gi.repository import Hinawa

from hinawa_utils.bebob.bebob_unit import BebobUnit
from hinawa_utils.ta1394.audio import AvcAudio
from hinawa_utils.ta1394.general import AvcConnection

__all__ = ['FocusriteSaffireproIoUnit']


class FocusriteSaffireproIoUnit(BebobUnit):
    _BASE_ADDR = 0x000100000000

    _INPUTS = (
        'Analog-1/2',
        'Analog-3/4',
        'Analog-5/6',
        'Analog-7/8',
        'S/PDIF-1/2',
        'ADAT-1/2',
        'ADAT-3/4',
        'ADAT-5/6',
        'ADAT-7/8',
        'ADAT-9/10',
        'ADAT-11/12',
        'ADAT-13/14',
        'ADAT-15/16',
    )

    _OUTPUTS = (
        'Analog-1/2',
        'Analog-3/4',
        'Analog-5/6',
        'Analog-7/8',
        'S/PDIF-1/2',
    )

    _MIXER_INPUT_OFFSETS = {
        # input pair    offset
        'Analog-1/2':   0x00,
        'Analog-3/4':   0x10,
        'Analog-5/6':   0x20,
        'Analog-7/8':   0x30,
        'S/PDIF-1/2':   0x40,
        'ADAT-1/2':     0x50,
        'ADAT-3/4':     0x60,
        'ADAT-5/6':     0x70,
        'ADAT-7/8':     0x80,
        'ADAT-9/10':    0x90,
        'ADAT-11/12':   0xa0,
        'ADAT-13/14':   0xb0,
        'ADAT-15/16':   0xc0,
    }

    _OUTPUT_SRC_OFFSETS = {
        # output pair   offset  pair_count
        'Analog-1/2':   (0xd0,  2),
        'Analog-3/4':   (0xe0,  3),
        'Analog-5/6':   (0xf8,  3),
        'Analog-7/8':   (0x110, 3),
        'S/PDIF-1/2':   (0x128, 3),
    }

    _OUTPUT_PARAMS_OFFSETS = {
        # output pair   offset
        'Analog-1/2':   0x140,
        'Analog-3/4':   0x144,
        'Analog-5/6':   0x148,
        'Analog-7/8':   0x14c,
    }

    _RATE_MODES = {
        'low':      (44100,  48000),
        'middle':   (88200,  96000),
        'high':     (176400, 192000),
    }

    _RATE_BITS = {
        44100:  1,
        48000:  2,
        88200:  3,
        96000:  4,
        176400: 5,
        192000: 6,
    }

    _CLOCK_BITS = {
        'Internal':     0,
        'S/PDIF':       2,
        'ADAT1':        3,
        'ADAT2':        4,
        'Word-clock':   5,
    }

    _IO10_CAPS = {
        'rate-modes':   ('low', 'middle'),
        'clocks':       ('Internal', 'S/PDIF'),
    }

    _IO26_CAPS = {
        'rate-modes':   ('low', 'middle', 'high'),
        'clocks':       ('Internal', 'S/PDIF', 'ADAT1', 'ADAT2', 'Word-clock'),
    }

    def __init__(self, path):
        CAPS = {
            0x000006: self._IO10_CAPS,
            0x000003: self._IO26_CAPS,
        }
        super().__init__(path)
        if self.model_id not in CAPS:
            raise OSError('Unsupported unit.')
        self._caps = CAPS[self.model_id]

    def _write_quads(self, offset, quads):
        frames = []
        count = len(quads)
        for i in range(count):
            frames.extend(pack('>I', quads[0]))
            quads = quads[1:]
        req = Hinawa.FwReq()
        if len(frames) == 4:
            tcode = Hinawa.FwTcode.WRITE_QUADLET_REQUEST
        else:
            tcode = Hinawa.FwTcode.WRITE_BLOCK_REQUEST
        req.transaction(self.get_node(), tcode,
                        self._BASE_ADDR + offset, len(frames), frames)

    def _read_quads(self, offset, count):
        quads = []
        req = Hinawa.FwReq()
        size = count * 4
        if size == 4:
            tcode = Hinawa.FwTcode.READ_QUADLET_REQUEST
        else:
            tcode = Hinawa.FwTcode.READ_BLOCK_REQUEST
        frames = bytearray(size)
        frames = req.transaction(self.get_node(), tcode,
                                 self._BASE_ADDR + offset, size, frames)
        for i in range(count):
            quads.append(unpack('>I', frames[0:4])[0])
            frames = frames[4:]
        return quads

    def get_mixer_input_labels(self):
        return self._INPUTS

    def set_mixer_input_balance(self, target, ch, balance):
        if target not in self._INPUTS:
            raise ValueError('Invalid value for input target.')
        offset = self._MIXER_INPUT_OFFSETS[target]
        quads = self._read_quads(offset, 4)
        pos = ch - 1
        total = quads[pos] + quads[pos + 2]
        quads[pos] = int(total * balance // 100)
        quads[pos + 2] = int(total * (100 - balance) // 100)
        self._write_quads(offset, quads)

    def get_mixer_input_balance(self, target, ch):
        if target not in self._INPUTS:
            raise ValueError('Invalid value for input target.')
        offset = self._MIXER_INPUT_OFFSETS[target]
        quads = self._read_quads(offset, 4)
        pos = ch - 1
        total = quads[pos] + quads[pos + 2]
        return quads[pos] * 100 // total

    def set_mixer_input_gain(self, target, ch, db):
        if target not in self._INPUTS:
            raise ValueError('Invalid value for input target.')
        offset = self._MIXER_INPUT_OFFSETS[target]
        quads = self._read_quads(offset, 4)
        pos = ch - 1
        total = quads[pos] + quads[pos + 2]
        data = AvcAudio.build_data_from_db(db)
        value = unpack('>H', data)[0]
        quads[pos] = int(value * quads[pos] // total)
        quads[pos + 2] = int(value * quads[pos + 2] // total)
        self._write_quads(offset, quads)

    def get_mixer_input_gain(self, target, ch):
        if target not in self._INPUTS:
            raise ValueError('Invalid value for input target.')
        offset = self._MIXER_INPUT_OFFSETS[target]
        quads = self._read_quads(offset, 4)
        pos = ch - 1
        total = quads[pos] + quads[pos + 2]
        data = pack('>H', total)
        return AvcAudio.parse_data_to_db(data)

    def get_output_destination_labels(self):
        return self._OUTPUTS

    def get_output_source_labels(self, target):
        labels = []
        labels.append('Stream-1/2')
        if target != 'Analog-1/2':
            labels.append(target.replace('Analog-', 'Stream-'))
        labels.append('Mixer-1/2')
        return labels

    def set_output_source(self, target, source):
        if target not in self._OUTPUTS:
            raise ValueError('Invalid value for output target.')
        labels = self.get_output_source_labels(target)
        if source not in labels:
            raise ValueError('Invalid value for target source.')
        offset, pair_count = self._OUTPUT_SRC_OFFSETS[target]
        quads = []
        for i in range(pair_count * 2):
            quads.append(0x00000000)
        quads[labels.index(source)] = 0x00007fff
        quads[len(labels) + labels.index(source)] = 0x00007fff
        self._write_quads(offset, quads)

    def get_output_source(self, target):
        if target not in self._OUTPUTS:
            raise ValueError('Invalid value for output target.')
        labels = self.get_output_source_labels(target)
        offset, pair_count = self._OUTPUT_SRC_OFFSETS[target]
        quads = self._read_quads(offset, pair_count * 2)
        for i in range(pair_count):
            if quads[i] > 0 and quads[pair_count + i] > 0:
                return labels[i]
        return None

    def get_output_labels(self):
        return list(self._OUTPUT_PARAMS_OFFSETS.keys())

    def _write_output_quad(self, target, quads):
        if target not in self._OUTPUT_PARAMS_OFFSETS:
            raise ValueError('Invalid value for output target.')
        offset = self._OUTPUT_PARAMS_OFFSETS[target]
        self._write_quads(offset, quads)

    def _read_output_quad(self, target):
        if target not in self._OUTPUT_PARAMS_OFFSETS:
            raise ValueError('Invalid value for output target.')
        offset = self._OUTPUT_PARAMS_OFFSETS[target]
        return self._read_quads(offset, 1)

    def set_output_volume(self, target, db):
        quads = self._read_output_quad(target)
        data = AvcAudio.build_data_from_db(db)
        quads[0] &= ~0x0000ffff
        quads[0] |= unpack('>H', data)[0]
        self._write_output_quad(target, quads)

    def get_output_volume(self, target):
        quads = self._read_output_quad(target)
        data = pack('>I', quads[0])
        return AvcAudio.parse_data_to_db(data[2:])

    def set_output_mute(self, target, enable):
        quads = self._read_output_quad(target)
        if enable:
            quads[0] |= 1 << 24
        else:
            quads[0] &= ~(1 << 24)
        self._write_output_quad(target, quads)

    def get_output_mute(self, target):
        quads = self._read_output_quad(target)
        return bool(quads[0] & (1 << 24))

    def set_output_hwctl(self, target, enable):
        quads = self._read_output_quad(target)
        if enable:
            quads[0] |= 1 << 26
        else:
            quads[0] &= ~(1 << 26)
        self._write_output_quad(target, quads)

    def get_output_hwctl(self, target):
        quads = self._read_output_quad(target)
        return bool(quads[0] & (1 << 26))

    def set_output_pad(self, target, enable):
        quads = self._read_output_quad(target)
        if enable:
            quads[0] |= 1 << 27
        else:
            quads[0] &= ~(1 << 27)
        self._write_output_quad(target, quads)

    def get_output_pad(self, target):
        quads = self._read_output_quad(target)
        return bool(quads[0] & (1 << 27))

    def set_output_dim(self, target, enable):
        quads = self._read_output_quad(target)
        if enable:
            quads[0] |= 1 << 28
        else:
            quads[0] &= ~(1 << 28)
        self._write_output_quad(target, quads)

    def get_output_dim(self, target):
        quads = self._read_output_quad(target)
        return bool(quads[0] & (1 << 28))

    def get_supported_rate_modes(self):
        return self._caps['rate-modes']

    def set_rate_mode(self, mode):
        if self.get_property('streaming'):
            raise OSError('Packet streaming starts')
        if mode not in self._caps['rate-modes']:
            raise ValueError('Invalid argument for frequency mode.')
        quads = []
        quads.append(self._RATE_BITS[self._RATE_MODES[mode][0]])
        # This corresponds to bus reset and the unit disappears from the bus,
        # then appears with a set mode.
        self._write_quads(0x150, quads)

    def get_rate_mode(self):
        rate = AvcConnection.get_plug_signal_format(self.fcp, 'input', 0)
        for mode, rates in self._RATE_MODES.items():
            if rate in rates and mode in self._caps['rate-modes']:
                break
        else:
            raise OSError('Invalid state for sampling rate.')
        return mode

    def get_supported_sampling_rates(self):
        rate = AvcConnection.get_plug_signal_format(self.fcp, 'input', 0)
        for mode, rates in self._RATE_MODES.items():
            if rate in rates and mode in self._caps['rate-modes']:
                break
        else:
            raise OSError('Invalid state for sampling rate.')
        return rates

    def set_sampling_rate(self, rate):
        if self.get_property('streaming'):
            raise OSError('Packet streaming starts')
        curr_rate = AvcConnection.get_plug_signal_format(self.fcp, 'input', 0)
        for mode, rates in self._RATE_MODES.items():
            if curr_rate in rates:
                break
        else:
            raise OSError('Invalid state for sampling rate.')
        if rate not in rates:
            raise ValueError('Invalid argument for sampling rate.')
        AvcConnection.set_plug_signal_format(self.fcp, 'input', 0, rate)
        AvcConnection.set_plug_signal_format(self.fcp, 'output', 0, rate)

    def get_sampling_rate(self):
        in_rate = AvcConnection.get_plug_signal_format(self.fcp, 'input', 0)
        out_rate = AvcConnection.get_plug_signal_format(self.fcp, 'output', 0)
        if in_rate != out_rate:
            raise OSError('Sampling rate mismatch: ', in_rate, out_rate)
        return in_rate

    def get_supported_clock_sources(self):
        return self._caps['clocks']

    def set_clock_source(self, source):
        if self.get_property('streaming'):
            raise OSError('Packet streaming starts')
        if source not in self._caps['clocks']:
            raise ValueError('Invalid argument for clock source.')
        quads = []
        quads.append(self._CLOCK_BITS[source])
        self._write_quads(0x0174, quads)

    def get_clock_source(self):
        quads = self._read_quads(0x0174, 1)
        quads[0] & 0xff
        for key, val in self._CLOCK_BITS.items():
            if val == quads[0] & 0xff:
                break
        else:
            raise OSError('Invalid state of clock source.')
        return key



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.dice.tcat_tcd22xx_spec import TcatTcd22xxSpec

__all__ = ['MaudioProfireSpec']


class MaudioProfireSpec(TcatTcd22xxSpec):
    MODELS = (
        (0x000d6c, 0x000010),   # Profire 2626
        (0x000d6c, 0x000011),   # Profire 610
    )

    _INPUTS = (
        (
            ('Analog',   'ins0',    8),
            ('ADAT-A',   'adat',    8),
            ('ADAT-B',   'adat',    8),
            ('S/PDIF',   'aes',     2),
        ),
        (
            ('Analog',   'ins0',    4),
            ('S/PDIF',   'aes',     2),
        ),
    )

    _OUTPUTS = (
        (
            ('Analog',   'ins0',    8),
            ('ADAT-A',   'adat',    8),
            ('ADAT-B',   'adat',    8),
            ('S/PDIF',   'aes',     2),
        ),
        (
            ('Analog',   'ins0',    8),
            ('S/PDIF',   'aes',     2),
        ),
    )

    _FIXED = (
        {},
        {
            0: ('src',  'ins0', 0),
            1: ('src',  'ins0', 1),
        },
    )



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from struct import pack, unpack
from pathlib import Path

from hinawa_utils.tscm.tscm_unit import TscmUnit

__all__ = ['TscmRackUnit']


class TscmRackUnit(TscmUnit):
    _OFFSET_CH_CTL = 0x0408

    _CH_LABELS = (
        'Analog-1', 'Analog-2', 'Analog-3', 'Analog-4',
        'Analog-5', 'Analog-6', 'Analog-7', 'Analog-8',
        'ADAT-1', 'ADAT-2', 'ADAT-3', 'ADAT-4',
        'ADAT-5', 'ADAT-6', 'ADAT-7', 'ADAT-8',
        'S/PDIF-1', 'S/PDIF-2',
    )
    _CH_FRAME_SIZE = 4

    def __init__(self, path):
        super().__init__(path)

        if self.model_name != 'FW-1804':
            raise ValueError('Unsupported model: {0}'.format(self.model_name))

        # For permanent cache.
        guid = self.get_property('guid')
        self._path = Path('/tmp/hinawa-{0:08x}'.format(guid))

        # For process local cache.
        self._cache = bytearray(len(self._CH_LABELS) * self._CH_FRAME_SIZE)

        # Sync these caches.
        self._load_cache()
        self._write_cache()

    def _load_cache(self):
        if not self._path.exists() or not self._path.is_file():
            # This is initial values.
            for i in range(len(self._CH_LABELS)):
                pos = i * self._CH_FRAME_SIZE
                self._cache[pos] = i
                if i % 2 == 0:
                    self._cache[pos + 1] = 0x00
                else:
                    self._cache[pos + 1] = 0xff
                self._cache[pos + 2] = 0x7f
                self._cache[pos + 3] = 0xff
        else:
            with self._path.open(mode='r') as f:
                for i, line in enumerate(f):
                    self._cache[i] = int(line.strip(), base=16)

        for i in range(len(self._CH_LABELS)):
            pos = i * self._CH_FRAME_SIZE
            self.write_quadlet(0x0408, self._cache[pos:pos + 4])

    def _write_cache(self):
        with self._path.open(mode='w+') as fd:
            for i, frame in enumerate(self._cache):
                fd.write('{0:02x}\n'.format(frame))

    def _write_frames(self, frames):
        # Write to the unit.
        self.write_quadlet(self._OFFSET_CH_CTL, frames)

        # Refresh process cache.
        ch = frames[0] & 0x7f
        for i, frame in enumerate(frames):
            pos = ch * self._CH_FRAME_SIZE
            self._cache[pos + i] = frame

        # Refresh permanent cache.
        self._write_cache()

    def _get_frames(self, ch):
        if ch not in self._CH_LABELS:
            raise ValueError(
                'Invalid argument for channel label: {0}'.format(ch))
        pos = self._CH_LABELS.index(ch) * self._CH_FRAME_SIZE
        return self._cache[pos:pos + self._CH_FRAME_SIZE]

    def get_channel_labels(self):
        return self._CH_LABELS

    def set_mute(self, ch, state):
        frames = self._get_frames(ch)
        if state:
            frames[0] |= 0x80
        else:
            frames[0] &= ~0x80
        self._write_frames(frames)

    def get_mute(self, ch):
        frames = self._get_frames(ch)
        return bool(frames[0] & 0x80)

    def set_balance(self, ch, balance):
        if balance < 0 or balance > 99:
            raise ValueError(
                'Invalid argument for LR Balance: {0}'.format(balance))
        frames = self._get_frames(ch)
        frames[1] = balance * 0xff // 99
        self._write_frames(frames)

    def get_balance(self, ch):
        frames = self._get_frames(ch)
        return frames[1] * 99 // 0xff

    def set_gain(self, ch, gain):
        if gain < 0 or gain > 99:
            raise ValueError('Invalid argument for gain: {0}'.format(gain))
        frames = self._get_frames(ch)
        gain = int(gain * 0x7fff // 99)
        data = pack('>H', gain)
        frames[2] = data[0]
        frames[3] = data[1]
        self._write_frames(frames)

    def get_gain(self, ch):
        frames = self._get_frames(ch)
        return unpack('>H', frames[2:4])[0] * 99 // 0x7fff



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from threading import Timer

import gi
gi.require_version('Hinawa', '3.0')
from gi.repository import Hinawa

from hinawa_utils.dice.dice_unit import DiceUnit

from hinawa_utils.dice.tcat_protocol_extension import ExtCtlSpace, ExtCapsSpace, ExtCmdSpace, ExtMixerSpace, ExtNewRouterSpace, ExtPeakSpace, ExtCurrentConfigSpace, ExtStandaloneSpace

from hinawa_utils.dice.tcat_tcd22xx_spec import TcatTcd22xxSpec
from hinawa_utils.dice.maudio_profire_spec import MaudioProfireSpec
from hinawa_utils.dice.focusrite_saffirepro_spec import FocusriteSaffireproSpec
from hinawa_utils.dice.presonus_firestudio_spec import PresonusFirestudioSpec

__all__ = ['DiceExtendedUnit']


class DiceExtendedUnit(DiceUnit):
    _RATE_MODES = {
        'low':      (32000, 48000),
        'middle':   (88200, 96000),
        'high':     (176400, 192000),
    }

    _SPECS = (
        MaudioProfireSpec,
        FocusriteSaffireproSpec,
        PresonusFirestudioSpec,
    )

    def __init__(self, fullpath):
        super().__init__(fullpath)

        req = Hinawa.FwReq()
        ExtCtlSpace.detect_layout(self._protocol, req)
        ExtCapsSpace.detect_caps(self._protocol, req)

        id_pair = (self.vendor_id, self.model_id)
        for spec in self._SPECS:
            if spec and id_pair in spec.MODELS:
                index = spec.MODELS.index(id_pair)
                break
        else:
            spec = TcatTcd22xxSpec
            index = 0
        self._spec = spec(index)

        # Cache current format of packets in data stream.
        self._cache_router_nodes()
        self.connect('notified', self._handle_notification)

    def _get_rate_mode(self, rate):
        for mode, rates in self._RATE_MODES.items():
            if rates[0] <= rate and rate <= rates[1]:
                return mode
        else:
            raise ValueError('Invalid argument for sampling rate.')

    def _handle_notification(self, obj, message):
        # MEMO: don't stop event loop.
        Timer(0, self._cache_router_nodes)

    def _cache_router_nodes(self):
        req = Hinawa.FwReq()

        rate = self._protocol.read_sampling_rate(req)
        mode = self._get_rate_mode(rate)

        entries = \
            ExtCurrentConfigSpace.read_router_config(self._protocol, req, mode)
        srcs, dsts = self._spec.get_available_ports(self._protocol, req, mode)

        routes = self._spec.normalize_router_entries(self._protocol, entries,
                                                     srcs, dsts)

        # MEMO: if registered entries are not generated by this module, update
        # them. Not friendly to the other programs while these entries are
        # valid for the programs.
        if entries != routes:
            ExtNewRouterSpace.set_entries(self._protocol, req, routes)
            ExtCmdSpace.initiate(self._protocol, req, 'load-from-router', mode)

        self._srcs = srcs
        self._dsts = dsts
        self._routes = routes

    def get_caps(self, category):
        if category not in self._protocol._ext_caps:
            raise ValueError('Invalid argument for capabilities.')
        return self._protocol._ext_caps[category]

    def get_stream_params(self, rate):
        if rate not in self._protocol.get_supported_sampling_rates():
            raise ValueError('Invalid argument for sampling rate.')
        mode = self._get_rate_mode(rate)
        req = Hinawa.FwReq()
        return ExtCurrentConfigSpace.read_stream_config(self._protocol, req, mode)

    def get_router_entries(self, rate):
        if rate not in self._protocol.get_supported_sampling_rates():
            raise ValueError('Invalid argument for sampling rate.')
        mode = self._get_rate_mode(rate)
        entries = []
        req = Hinawa.FwReq()
        routes = ExtCurrentConfigSpace.read_router_config(self._protocol, req,
                                                          mode)
        for route in routes:
            for src in self._srcs:
                if route['src-blk'] == src[1] and route['src-ch'] in src[2]:
                    break
            else:
                continue
            for dst in self._dsts:
                if route['dst-blk'] == dst[1] and route['dst-ch'] in dst[2]:
                    break
            else:
                continue
            entry = {
                'src': '{0}:{1}'.format(src[0], src[2].index(route['src-ch'])),
                'dst': '{0}:{1}'.format(dst[0], dst[2].index(route['dst-ch'])),
            }
            entries.append(entry)
        return entries

    def store_to_storage(self):
        if not self._protocol._ext_caps['general']['storage-available']:
            raise RuntimeError('This feature is not supported.')

        categories = []
        if self._protocol._ext_caps['general']['storable-stream-conf']:
            categories.append('stream-config')
        if self._protocol._ext_caps['mixer']['is-storable']:
            categories.append('mixer')
        if self._protocol._ext_caps['router']['is-storable']:
            categories.append('router')
        if len(categories) == 0:
            raise RuntimeError('Nothing can be stored.')

        req = Hinawa.FwReq()
        rate = self._protocol.read_sampling_rate(req)
        mode = self._get_rate_mode(rate)
        ExtCmdSpace(self._protocol, req, 'load-to-storage', mode)

        # MEMO: however, in most models, configuration of router is stored by
        # 'load-from-router' command.
        return categories

    def load_from_storage(self):
        if not self._protocol._ext_caps['general']['storage-available']:
            raise RuntimeError('This feature is not supported.')

        categories = []
        if self._protocol._ext_caps['general']['storable-stream-conf']:
            categories.append('stream-config')
        if self._protocol._ext_caps['mixer']['is-storable']:
            categories.append('mixer')
        if self._protocol._ext_caps['router']['is-storable']:
            categories.append('router')
        if len(categories) == 0:
            raise RuntimeError('Nothing can be loaded.')

        req = Hinawa.FwReq()
        rate = self._protocol.read_sampling_rate(req)
        mode = self._get_rate_mode(rate)
        ExtCmdSpace.initiate(self._protocol, req, 'load-from-storage', mode)
        # MEMO: I expect notification here.
        return categories

    def _find_route_pairs(self, target):
        for dst in self._dsts:
            if target == dst[0]:
                break
        else:
            raise ValueError('Invalid argument for destination.')

        pairs = []
        for route in self._routes:
            if route['dst-blk'] == dst[1] and route['dst-ch'] in dst[2]:
                pairs.append(route)

        return sorted(pairs, key=lambda pair: (pair['src-ch'], pair['dst-ch']))

    def _set_target_source(self, target, source):
        pairs = self._find_route_pairs(target)

        if source == 'None':
            indices = []
            for pair in pairs:
                indices.append(self._routes.index(pair))
            # Pop backwards.
            indices.sort(reverse=True)
            for index in indices:
                self._routes.pop(index)
        else:
            for dst in self._dsts:
                if target == dst[0]:
                    break
            for src in self._srcs:
                if source == src[0]:
                    break

            if len(pairs) > 0:
                # Left->Left, Right->Right.
                for i, pair in enumerate(pairs):
                    pair['src-blk'] = src[1]
                    pair['src-ch'] = src[2][i]
            else:
                for i in range(2):
                    pair = {
                        'src-blk':  src[1],
                        'src-ch':   src[2][i],
                        'dst-blk':  dst[1],
                        'dst-ch':   dst[2][i],
                        'peak':     0,
                    }
                    self._routes.append(pair)

        req = Hinawa.FwReq()
        rate = self._protocol.read_sampling_rate(req)
        mode = self._get_rate_mode(rate)
        ExtNewRouterSpace.set_entries(self._protocol, req, self._routes)
        ExtCmdSpace.initiate(self._protocol, req, 'load-from-router', mode)

    def _get_target_source(self, target):
        pairs = self._find_route_pairs(target)

        for pair in pairs:
            for src in self._srcs:
                if pair['src-blk'] == src[1] and pair['src-ch'] in src[2]:
                    return src[0]
        return 'None'

    def get_output_labels(self):
        labels = []
        for dst in self._dsts:
            if dst[1] not in ('mixer-tx0', 'mixer-tx1', 'avs0', 'avs1'):
                labels.append(dst[0])
        return labels

    def get_output_source_labels(self):
        labels = ['None']
        for src in self._srcs:
            labels.append(src[0])
        return labels

    def set_output_source(self, target, source):
        if target not in self.get_output_labels():
            raise ValueError('Invalid argument for output pair.')
        if source not in self.get_output_source_labels():
            raise ValueError('Invalid argument for output source pair.')
        self._set_target_source(target, source)

    def get_output_source(self, target):
        if target not in self.get_output_labels():
            raise ValueError('Invalid argument for output pair.')
        return self._get_target_source(target)

    def get_tx_stream_labels(self):
        labels = []
        for dst in self._dsts:
            if dst[1] in ('avs0', 'avs1'):
                labels.append(dst[0])
        return labels

    def get_tx_stream_source_labels(self):
        labels = ['None']
        for src in self._srcs:
            labels.append(src[0])
        return labels

    def set_tx_stream_source(self, target, source):
        if target not in self.get_tx_stream_labels():
            raise ValueError('Invalid argument for tx stream.')
        if source not in self.get_tx_stream_source_labels():
            raise ValueError('Invalid argument for source of tx stream.')
        self._set_target_source(target, source)

    def get_tx_stream_source(self, target):
        if target not in self.get_tx_stream_labels():
            raise ValueError('Invalid argument for tx stream.')
        return self._get_target_source(target)

    def get_mixer_output_labels(self):
        labels = []
        for src in self._srcs:
            if src[1] == 'mixer':
                labels.append(src[0])
        return labels

    def get_mixer_input_labels(self):
        labels = []
        for dst in self._dsts:
            if dst[1] in ('mixer-tx0', 'mixer-tx1'):
                labels.append(dst[0])
        return labels

    def get_mixer_source_labels(self):
        labels = ['None']
        for src in self._srcs:
            if src[1] != 'mixer':
                labels.append(src[0])
        return labels

    def set_mixer_source(self, target, source):
        if target not in self.get_mixer_input_labels():
            raise ValueError('Invalid argument for mixer pair.')
        if source not in self.get_mixer_source_labels():
            raise ValueError('Invalid argument for mixer source pair.')
        self._set_target_source(target, source)

    def get_mixer_source(self, target):
        if target not in self.get_mixer_input_labels():
            raise ValueError('Invalid argument for mixer pair.')
        return self._get_target_source(target)

    def _get_mixer_gains(self, req, output, input, ch):
        if self.get_mixer_source(input) == 'None':
            raise ValueError('This input to mixer has no source.')
        if output not in self.get_mixer_output_labels():
            raise ValueError('Invalid argument for mixer stereo pair.')
        if input not in self.get_mixer_input_labels():
            raise ValueError('Invalid argument for mixer input stereo pair.')
        if ch not in (0, 1):
            raise ValueError('Invalid argument for channel in stereo pair.')

        for dst in self._dsts:
            if dst[0] == output:
                break
        for src in self._srcs:
            if src[0] == input:
                break

        gains = []
        total = 0
        src_ch = src[2][ch]
        for dst_ch in dst[2]:
            val = ExtMixerSpace.read_gain(self._protocol, req, dst_ch, src_ch)
            gain = {
                'dst-ch':   dst_ch,
                'src-ch':   src_ch,
                'val':      val,
            }
            gains.append(gain)
            total += val

        # normalize.
        if total > ExtMixerSpace.MAX_COEFF:
            gains[0]['val'] = gains[0]['val'] * \
                ExtMixerSpace.MAX_COEFF // total
            gains[1]['val'] = gains[1]['val'] * \
                ExtMixerSpace.MAX_COEFF // total

        return gains

    def set_mixer_gain(self, output, input, ch, db):
        req = Hinawa.FwReq()
        gains = self._get_mixer_gains(req, output, input, ch)
        total = gains[0]['val'] + gains[1]['val']
        val = ExtMixerSpace.build_val_from_db(db)
        if total == 0:
            gains[ch] = val
            gains[(ch + 1) % 2] = 0
        else:
            gains[0]['val'] = gains[0]['val'] * val // total
            gains[1]['val'] = gains[1]['val'] * val // total
        for gain in gains:
            ExtMixerSpace.write_gain(self._protocol, req, gain['dst-ch'],
                                     gain['src-ch'], gain['val'])

    def get_mixer_gain(self, output, input, ch):
        req = Hinawa.FwReq()
        gains = self._get_mixer_gains(req, output, input, ch)
        total = gains[0]['val'] + gains[1]['val']
        return ExtMixerSpace.parse_val_to_db(total)

    def set_mixer_balance(self, output, input, ch, balance):
        req = Hinawa.FwReq()
        gains = self._get_mixer_gains(req, output, input, ch)
        total = gains[0]['val'] + gains[1]['val']
        gains[0]['val'] = int(total * (100 - balance) // 100)
        gains[1]['val'] = total - gains[0]['val']
        for gain in gains:
            ExtMixerSpace.write_gain(self._protocol, req, gain['dst-ch'],
                                     gain['src-ch'], gain['val'])

    def get_mixer_balance(self, output, input, ch):
        req = Hinawa.FwReq()
        gains = self._get_mixer_gains(req, output, input, ch)
        total = gains[0]['val'] + gains[1]['val']
        if total == 0:
            balance = ch * 100.0
        else:
            balance = float(100 * gains[1]['val'] // total)
        return balance

    def get_mixer_saturations(self):
        outputs = self.get_mixer_output_labels()

        req = Hinawa.FwReq()
        rate = self._protocol.read_sampling_rate(req)
        mode = self._get_rate_mode(rate)
        saturations = ExtMixerSpace.read_saturation(self._protocol, req, mode)

        mixer_saturations = {}
        for i, saturation in enumerate(saturations):
            index = i // 2
            label = outputs[index]
            if label not in mixer_saturations:
                mixer_saturations[label] = [False, False]
            mixer_saturations[label][i % 2] = saturation
        return mixer_saturations

    def get_metering(self):
        meters = {}

        req = Hinawa.FwReq()
        for peak in ExtPeakSpace.get(self._protocol, req):
            for src in self._srcs:
                if peak['src-blk'] == src[1] and peak['src-ch'] in src[2]:
                    break
            else:
                continue
            for dst in self._dsts:
                if peak['dst-blk'] == dst[1] and peak['dst-ch'] in dst[2]:
                    break
            else:
                continue

            src_index = src[2].index(peak['src-ch'])
            dst_index = dst[2].index(peak['dst-ch'])

            if src[0] not in meters:
                meters[src[0]] = {0: {}, 1: {}}
            if dst[0] not in meters[src[0]][src_index]:
                meters[src[0]][src_index][dst[0]] = {0: 0, 1: 0}
            meters[src[0]][src_index][dst[0]][dst_index] = peak['peak']

        return meters

    def set_standalone_clock_source(self, source):
        req = Hinawa.FwReq()
        labels = self._protocol.get_clock_source_names()
        if source not in labels or source == 'Unused':
            raise ValueError('Invalid argument for clock source.')
        alias = self._protocol.CLOCK_BITS[labels.index(source)]
        ExtStandaloneSpace.write_clock_source(self._protocol, req, alias)

    def get_standalone_clock_source(self):
        req = Hinawa.FwReq()
        labels = self._protocol.get_clock_source_names()
        src = ExtStandaloneSpace.read_clock_source(self._protocol, req)
        index = {v: k for k, v in self._protocol.CLOCK_BITS.items()}[src]
        return labels[index]

    def get_standalone_clock_source_param_options(self, source):
        labels = self._protocol.get_clock_source_names()
        if source not in labels or source == 'Unused':
            raise ValueError('Invalid argument for clock source.')
        alias = self._protocol.CLOCK_BITS[labels.index(source)]

        return ExtStandaloneSpace.get_source_param_options(self._protocol, alias)

    def set_standalone_clock_source_params(self, source, params):
        labels = self._protocol.get_clock_source_names()
        if source not in labels or source == 'Unused':
            raise ValueError('Invalid argument for clock source.')
        alias = self._protocol.CLOCK_BITS[labels.index(source)]

        param_options = \
            ExtStandaloneSpace.get_source_param_options(self._protocol, alias)
        for name, options in param_options.items():
            if name not in params:
                raise ValueError('Invalid argument for params.')

        req = Hinawa.FwReq()
        ExtStandaloneSpace.write_clock_source_params(self._protocol, req, alias,
                                                     params)

    def get_standalone_clock_source_params(self, source):
        labels = self._protocol.get_clock_source_names()
        if source not in labels or source == 'Unused':
            raise ValueError('Invalid argument for clock source.')
        alias = self._protocol.CLOCK_BITS[labels.index(source)]

        req = Hinawa.FwReq()
        return ExtStandaloneSpace.read_clock_source_params(self._protocol, req,
                                                           alias)



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from struct import unpack, pack
from pathlib import Path

import gi
gi.require_version('Hinawa', '3.0')
from gi.repository import Hinawa

from hinawa_utils.bebob.maudio_protocol_abstract import MaudioProtocolAbstract

from hinawa_utils.ta1394.general import AvcConnection
from hinawa_utils.ta1394.audio import AvcAudio

__all__ = ['MaudioProtocolSpecial']


class MaudioProtocolSpecial(MaudioProtocolAbstract):
    BASE_ADDR = 0xffc700700000
    __IDS = (
        0x010071,   # Firewire 1814
        0x010091,   # ProjectMix I/O
    )

    __INPUT_LABELS = (
        'stream-1/2', 'stream-3/4',
        'analog-1/2', 'analog-3/4', 'analog-5/6', 'analog-7/8',
        'spdif-1/2', 'adat-1/2', 'adat-3/4', 'adat-5/6', 'adat-7/8',
    )

    __OUTPUT_LABELS = ('analog-1/2', 'analog-3/4')

    __HP_LABELS = ('headphone-1/2', 'headphone-3/4')

    __MIXER_LABELS = ('mixer-1/2', 'mixer-3/4')

    __HP_SOURCE_LABELS = ('mixer-1/2', 'mixer-3/4', 'aux-1/2')

    __METERING_LABELS = (
        'analog-in-1', 'analog-in-2', 'analog-in-3', 'analog-in-4',
        'analog-in-5', 'analog-in-6', 'analog-in-7', 'analog-in-8',
        'spdif-in-1', 'spdif-in-2',
        'adat-in-1', 'adat-in-2', 'adat-in-3', 'adat-in-4',
        'adat-in-5', 'adat-in-6', 'adat-in-7', 'adat-in-8',
        'analog-out-1', 'analog-out-2',
        'analog-out-3', 'analog-out-4',
        'spdif-out-1', 'spdif-out-2',
        'adat-out-1', 'adat-out-2', 'adat-out-3', 'adat-out-4',
        'adat-out-5', 'adat-out-6', 'adat-out-7', 'adat-out-8',
        'headphone-out-1', 'headphone-out-2',
        'headphone-out-3', 'headphone-out-4',
        'aux-out-1', 'aux-out-2')

    # LR balance: 0x40-60
    # Write 32 bit, upper 16bit for left channel and lower 16bit for right.
    # The value is between 0x800(L) to 0x7FFE(R) as the same as '10.3.3 LR
    # Balance Control' in 'AV/C Audio Subunit Specification 1.0 (1394TA
    # 1999008)'.

    def __init__(self, unit, debug):
        if unit.model_id not in self.__IDS:
            raise OSError('Not supported')

        super().__init__(unit, debug)

        # For process local cache.
        self._cache = bytearray(160)
        # For permanent cache.
        guid = self.unit.get_property('guid')
        self.__path = Path('/tmp/hinawa-{0:08x}'.format(guid))
        self.__load_cache()

    # Read transactions are not allowed. We cache data.
    def __load_cache(self):
        if not self.__path.exists() or not self.__path.is_file():
            # This is initial value.
            cache = [
                0x00, 0x00, 0x00, 0x00,  # gain of inputs from stream 1/2
                0x00, 0x00, 0x00, 0x00,  # gain of inputs from stream 3/4
                0x00, 0x00, 0x00, 0x00,  # volume of outputs to analog 1/2
                0x00, 0x00, 0x00, 0x00,  # volume of outputs to analog 3/4
                0x00, 0x00, 0x00, 0x00,  # gain of inputs from analog 1/2
                0x00, 0x00, 0x00, 0x00,  # gain of inputs from analog 3/4
                0x00, 0x00, 0x00, 0x00,  # gain of inputs from analog 5/6
                0x00, 0x00, 0x00, 0x00,  # gain of inputs from analog 7/8
                0x00, 0x00, 0x00, 0x00,  # gain of inputs from spdif 1/2
                0x00, 0x00, 0x00, 0x00,  # gain of inputs from adat 1/2
                0x00, 0x00, 0x00, 0x00,  # gain of inputs from adat 3/4
                0x00, 0x00, 0x00, 0x00,  # gain of inputs from adat 5/6
                0x00, 0x00, 0x00, 0x00,  # gain of inputs from adat 7/8
                0x00, 0x00, 0x00, 0x00,  # volume of outputs to aux 1/2
                0x00, 0x00, 0x00, 0x00,  # volume of outputs to headphone 1/2
                0x00, 0x00, 0x00, 0x00,  # volume of outputs to headophone 3/4
                0x7F, 0xFE, 0x80, 0x00,  # balance of inputs from analog 1/2
                0x7F, 0xFE, 0x80, 0x00,  # balance of inputs from analog 3/4
                0x7F, 0xFE, 0x80, 0x00,  # balance of inputs from analog 5/6
                0x7F, 0xFE, 0x80, 0x00,  # balance of inputs from analog 7/8
                0x7F, 0xFE, 0x80, 0x00,  # balance of inputs from spdif 1/2
                0x7F, 0xFE, 0x80, 0x00,  # balance of inputs from adat 1/2
                0x7F, 0xFE, 0x80, 0x00,  # balance of inputs from adat 3/4
                0x7F, 0xFE, 0x80, 0x00,  # balance of inputs from adat 5/6
                0x7F, 0xFE, 0x80, 0x00,  # balance of inputs from adat 7/8
                0x80, 0x00, 0x80, 0x00,  # inputs of stream 1/2 to aux
                0x80, 0x00, 0x80, 0x00,  # inputs of stream 3/4 to aux
                0x80, 0x00, 0x80, 0x00,  # inputs of analog 1/2 to aux
                0x80, 0x00, 0x80, 0x00,  # inputs of analog 3/4 to aux
                0x80, 0x00, 0x80, 0x00,  # inputs of analog 5/6 to aux
                0x80, 0x00, 0x80, 0x00,  # inputs of analog 7/8 to aux
                0x80, 0x00, 0x80, 0x00,  # inputs of spdif 1/2 to aux
                0x80, 0x00, 0x80, 0x00,  # inputs of adat 1/2 to aux
                0x80, 0x00, 0x80, 0x00,  # inputs of adat 3/4 to aux
                0x80, 0x00, 0x80, 0x00,  # inputs of adat 5/6 to aux
                0x80, 0x00, 0x80, 0x00,  # inputs of adat 7/8 to aux
                0x00, 0x00, 0x00, 0x00,  # inputs of analog/digital for mixer
                0x00, 0x00, 0x00, 0x09,  # inputs of stream for mixer
                0x00, 0x02, 0x00, 0x01,  # source for headphone out 1/2 and 3/4
                0x00, 0x00, 0x00, 0x00]  # source for analog out 1/2 and 3/4
        else:
            cache = bytearray(160)
            with self.__path.open(mode='r') as f:
                for i, line in enumerate(f):
                    cache[i] = int(line.strip(), base=16)

        self.__write_data(0, cache)

    def __write_data(self, offset, data):
        # Write to the unit.
        count = 0
        req = Hinawa.FwReq()
        while True:
            if len(data) == 4:
                tcode = Hinawa.FwTcode.WRITE_QUADLET_REQUEST
            else:
                tcode = Hinawa.FwTcode.WRITE_BLOCK_REQUEST
            try:
                req.transaction(self.unit.get_node(), tcode,
                                self.BASE_ADDR + offset, len(data), data)
                break
            except Exception:
                if count > 10:
                    raise OSError('Fail to communicate to the unit.')
                count += 1
        # Refresh process cache.
        for i, datum in enumerate(data):
            self._cache[offset + i] = datum
        # Refresh permanent cache.
        with self.__path.open(mode='w+') as fd:
            for i, datum in enumerate(self._cache):
                fd.write('{0:02x}\n'.format(datum))

    def __set_volume(self, offset, db):
        if offset > len(self._cache):
            raise ValueError('Invalid argument for offset on address space')
        data = AvcAudio.build_data_from_db(db)
        self.__write_data(offset, data)

    def __get_volume(self, offset):
        if offset > len(self._cache):
            raise ValueError('Invalid argument for offset on address space')
        return AvcAudio.parse_data_to_db(self._cache[offset:offset + 2])

    def get_input_labels(self):
        return self.__INPUT_LABELS

    def __get_input_gain_offset(self, target, ch):
        if target not in self.__INPUT_LABELS:
            raise ValueError('invalid argument for input stereo pair')
        if ch > 1:
            raise ValueError('Invalid argument for stereo pair channel')
        offset = self.__INPUT_LABELS.index(target) * 4
        if offset >= 8:
            offset += 8
        return offset + ch * 2

    def set_input_gain(self, target, ch, db):
        offset = self.__get_input_gain_offset(target, ch)
        self.__set_volume(offset, db)

    def get_input_gain(self, target, ch):
        offset = self.__get_input_gain_offset(target, ch)
        return self.__get_volume(offset)

    def get_input_balance_labels(self):
        labels = []
        for label in self.__INPUT_LABELS:
            if label.find('stream-') == 0:
                continue
            labels.append(label)
        return labels

    def set_input_balance(self, target, ch, balance):
        if target not in self.__INPUT_LABELS:
            raise ValueError('invalid argument for input stereo pair')
        offset = (self.__INPUT_LABELS.index(target) - 2 + 16) * 4 + ch * 2
        data = AvcAudio.build_data_from_db(balance)
        self.__write_data(offset, data)

    def get_input_balance(self, target, ch):
        if target not in self.__INPUT_LABELS:
            raise ValueError('invalid argument for input stereo pair')
        offset = (self.__INPUT_LABELS.index(target) - 2 + 16) * 4 + ch * 2
        data = self._cache[offset:offset + 2]
        return AvcAudio.parse_data_to_db(data)

    def get_output_labels(self):
        return self.__OUTPUT_LABELS

    def __set_output_volume_offset(self, target, ch):
        if target not in self.__OUTPUT_LABELS:
            raise ValueError('invalid argument for output stereo pair')
        if ch > 1:
            raise ValueError('Invalid argument for stereo pair channel')
        return 8 + self.__OUTPUT_LABELS.index(target) * 4 + ch * 2

    def set_output_volume(self, target, ch, value):
        offset = self.__set_output_volume_offset(target, ch)
        self.__set_volume(offset, value)

    def get_output_volume(self, target, ch):
        offset = self.__set_output_volume_offset(target, ch)
        return self.__get_volume(offset)

    def __get_aux_volume_offset(self, ch):
        if ch > 1:
            raise ValueError('Invalid argument for stereo pair channel')
        return 52 + ch * 2

    def set_aux_volume(self, ch, value):
        offset = self.__get_aux_volume_offset(ch)
        self.__set_volume(offset, value)

    def get_aux_volume(self, ch):
        offset = self.__get_aux_volume_offset(ch)
        return self.__get_volume(offset)

    def get_headphone_labels(self):
        return self.__HP_LABELS

    def __get_headphone_volume_offset(self, target, ch):
        if target not in self.__HP_LABELS:
            raise ValueError('invalid argument for heaphone stereo pair')
        if ch > 1:
            raise ValueError('Invalid argument for stereo pair channel')
        return 56 + self.__HP_LABELS.index(target) * 4 + ch * 2

    def set_headphone_volume(self, target, ch, value):
        offset = self.__get_headphone_volume_offset(target, ch)
        self.__set_volume(offset, value)

    def get_headphone_volume(self, target, ch):
        offset = self.__get_headphone_volume_offset(target, ch)
        return self.__get_volume(offset)

    def get_aux_input_labels(self):
        return self.__INPUT_LABELS

    def __get_aux_input_offset(self, target, ch):
        if target not in self.__INPUT_LABELS:
            raise ValueError('Invalid argument for input stereo pair')
        if ch > 1:
            raise ValueError('Invalid argument for stereo pair channel')
        return 104 + self.__INPUT_LABELS.index(target) * 4 + ch * 2

    def set_aux_input(self, target, ch, value):
        offset = self.__get_aux_input_offset(target, ch)
        self.__set_volume(offset, value)

    def get_aux_input(self, target, ch):
        offset = self.__get_aux_input_offset(target, ch)
        return self.__get_volume(offset)

    def get_mixer_labels(self):
        return self.__MIXER_LABELS

    def get_mixer_source_labels(self):
        return self.__INPUT_LABELS

    def __calculate_mixer_input_bit(self, mixer, source):
        if mixer not in self.__MIXER_LABELS:
            raise ValueError('invalid argument for mixer stereo pair')
        if source not in self.__INPUT_LABELS:
            raise ValueError('Invalid argument for source stereo pair')
        if source.find('stream') == 0:
            pos = 0
            if source == 'stream-3/4':
                pos = 2
            pos = pos + self.__MIXER_LABELS.index(mixer)
        else:
            if source.find('analog') == 0:
                pos = self.__MIXER_LABELS.index(mixer) * 4
                pos = pos + self.__INPUT_LABELS.index(source) - 2
            else:
                pos = 16 + (self.__INPUT_LABELS.index(source) - 6) * 2
                pos = pos + self.__MIXER_LABELS.index(mixer)
        return pos

    def __get_mixer_offset(self, source):
        if source.find('stream') == 0:
            offset = 148
        else:
            offset = 142
        return offset

    def set_mixer_routing(self, mixer, source, enable):
        pos = self.__calculate_mixer_input_bit(mixer, source)
        offset = self.__get_mixer_offset(source)
        val = unpack('>I', self._cache[offset:offset + 4])[0]
        if enable > 0:
            val |= (1 << pos)
        else:
            val &= ~(1 << pos)
        data = pack('>I', val)
        self.__write_data(offset, data)

    def get_mixer_routing(self, mixer, source):
        pos = self.__calculate_mixer_input_bit(mixer, source)
        offset = self.__get_mixer_offset(source)
        val = unpack('>I', self._cache[offset:offset + 4])[0]
        return bool(val & (1 << pos))

    def get_headphone_source_labels(self, target):
        return self.__HP_SOURCE_LABELS

    def set_headphone_source(self, target, source):
        if target not in self.__HP_LABELS:
            raise ValueError('Invalid argument for output stereo pair')
        if source not in self.__HP_SOURCE_LABELS:
            raise ValueError('Invalid argument for headphone source')
        index = (self.__HP_LABELS.index(target) + 1) % 2
        pos = self.__HP_SOURCE_LABELS.index(source)
        offset = 152
        vals = list(unpack('>2H', self._cache[offset:offset + 4]))
        vals[index] = 1 << pos
        data = pack('>2H', vals[0], vals[1])
        self.__write_data(offset, data)

    def get_headphone_source(self, target):
        if target not in self.__HP_LABELS:
            raise ValueError('Invalid argument for output stereo pair')
        index = (self.__HP_LABELS.index(target) + 1) % 2
        offset = 152
        vals = unpack('>2H', self._cache[offset:offset + 4])
        for i, source in enumerate(self.__HP_SOURCE_LABELS):
            if vals[index] & (1 << i):
                return source
        else:
            raise IOError('Unexpected state of register cache')

    def __calculate_output_source_bit(self, target):
        if target not in self.__OUTPUT_LABELS:
            raise ValueError('Invalid argument for output stereo pair')
        return self.__OUTPUT_LABELS.index(target)

    def get_output_source_labels(self, target):
        labels = []
        if target not in self.__OUTPUT_LABELS:
            raise ValueError('Invalid argument for output stereo pair')
        if target.find('1/2') > 0:
            labels.append('mixer-1/2')
        else:
            labels.append('mixer-3/4')
        labels.append('aux-1/2')
        return labels

    def set_output_source(self, target, source):
        if target not in self.__OUTPUT_LABELS:
            raise ValueError('Invalid argument for output stereo pair')
        pos = self.__OUTPUT_LABELS.index(target)
        labels = self.get_output_source_labels(target)
        if source not in labels:
            raise ValueError('Invalid argument for output source pair')
        offset = 156
        val = unpack('>I', self._cache[offset:offset + 4])[0]
        if labels.index(source) > 0:
            val |= 1 << pos
        else:
            val &= ~(1 << pos)
        data = pack('>I', val)
        self.__write_data(offset, data)

    def get_output_source(self, target):
        if target not in self.__OUTPUT_LABELS:
            raise ValueError('Invalid argument for output stereo pair')
        pos = self.__OUTPUT_LABELS.index(target)
        labels = self.get_output_source_labels(target)
        offset = 156
        val = unpack('>I', self._cache[offset:offset + 4])[0]
        if not val & (1 << pos):
            return labels[0]
        else:
            return labels[1]

    # 0x0000ffff - 0x7fffffff
    # db = 20 * log10(vol / 0x80000000)
    # vol = 0, then db = -144.0
    # may differs analog-in and the others.
    def get_meters(self):
        meters = {}
        req = Hinawa.FwReq()
        data = [0] * 84
        data = req.transaction(self.unit.get_node(),
                               Hinawa.FwTcode.READ_BLOCK_REQUEST,
                               self._ADDR_FOR_METERING, 84, data)
        meters['switch-0'] = data[0]
        meters['rotery-0'] = data[1]
        meters['rotery-1'] = data[2]
        meters['rotery-2'] = data[3]
        data = data[4:]
        for i, label in enumerate(self.__METERING_LABELS):
            meters[label] = unpack('>H', data[i * 2:(i + 1) * 2])[0]
        meters['rate'] = AvcConnection.SAMPLING_RATES[(data[-1] >> 8) & 0x0f]
        meters['sync'] = (data[-1] & 0x0f) > 0
        return meters

    def get_clock_source_labels(self):
        return ()

    def set_clock_source(self, src):
        print('Not supported. Please use ALSA control interface for this aim.')

    def get_clock_source(self):
        print('Not supported. Please use ALSA control interface for this aim.')

    def get_sampling_rate(self):
        # Current rate is correctly returned from a plug 0 for output direction
        # only.
        return AvcConnection.get_plug_signal_format(self.unit.fcp, 'output', 0)



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from threading import Thread
from math import log10
from struct import pack, unpack
from pathlib import Path

import gi
gi.require_version('GLib', '2.0')
gi.require_version('Hinawa', '3.0')
from gi.repository import GLib, Hinawa

from hinawa_utils.fireface.ff_config_rom_parser import FFConfigRomParser
from hinawa_utils.fireface.ff_option_reg import FFOptionReg
from hinawa_utils.fireface.ff_status_reg import FFStatusReg, FFClkLabels
from hinawa_utils.fireface.ff_mixer_reg import FFMixerRegs
from hinawa_utils.fireface.ff_out_reg import FFOutRegs

__all__ = ['FFUnit']


class FFUnit(Hinawa.SndUnit):
    __MODELS = {
        0x000001:   'Fireface800',
        0x000002:   'Fireface400',
    }
    __REGS = {
        # model_id: (option offset, mixer offset, out offset)
        0x000001:   (0x0000fc88f014, 0x000080080000, 0x000080081f80),
        0x000002:   (0x00008010051c, 0x000080080000, 0x000080080f80),
    }
    __SPECS = {
        0x000001: {
            'analog':   10,
            'spdif':    2,
            'adat':     16,
            'stream':   28,
            'avail':    32,
        },
        0x000002: {
            'analog':   8,
            'spdif':    2,
            'adat':     8,
            'stream':   18,
            'avail':    18,
        },
    }

    __MUTE_VAL = 0x00000000
    __ZERO_VAL = 0x00008000
    __MIN_VAL = 0x00000001
    __MAX_VAL = 0x00010000

    def __init__(self, path):
        super().__init__()
        self.open(path)

        ctx = GLib.MainContext.new()
        self.create_source().attach(ctx)
        self.__unit_dispatcher = GLib.MainLoop.new(ctx, False)
        self.__unit_th = Thread(target=lambda d: d.run(), args=(self.__unit_dispatcher, ))
        self.__unit_th.start()

        node = self.get_node()
        ctx = GLib.MainContext.new()
        node.create_source().attach(ctx)
        self.__node_dispatcher = GLib.MainLoop.new(ctx, False)
        self.__node_th = Thread(target=lambda d: d.run(), args=(self.__node_dispatcher, ))
        self.__node_th.start()

        parser = FFConfigRomParser()
        info = parser.parse_rom(self.get_node().get_config_rom())
        if info['model_id'] not in self.__MODELS:
            raise OSError('Unsupported model.')

        self.__name = self.__MODELS[info['model_id']]
        self.__regs = self.__REGS[info['model_id']]
        self.__spec = self.__SPECS[info['model_id']]

        guid = self.get_property('guid')
        self._path = Path('/tmp/hinawa-{0:08x}'.format(guid))

        if self._path.exists() and self._path.is_file():
            self.__read_cache_from_file()
        else:
            self.__option_cache = self.__create_option_initial_cache()
            self.__mixer_cache = self.__create_mixer_initial_cache()
            self.__out_cache = self.__create_out_initial_cache()
            self.__load_settings()
            self.__write_cache_to_file()

        self.__load_option_settings()

    def release(self):
        self.__unit_dispatcher.quit()
        self.__node_dispatcher.quit()
        self.__unit_th.join()
        self.__node_th.join()

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, trace):
        self.release()

    def __read_cache_from_file(self):
        self.__option_cache = []
        self.__mixer_cache = []
        self.__out_cache = []
        with self._path.open(mode='r') as f:
            for i, line in enumerate(f):
                line = line.strip()
                tokens = line.split(' ')
                reg_type = tokens[0]
                reg_val = int(tokens[1], 16)
                if reg_type == 'option':
                    self.__option_cache.append(reg_val)
                elif reg_type == 'mixer':
                    self.__mixer_cache.append(reg_val)
                elif reg_type == 'out':
                    self.__out_cache.append(reg_val)

    def __write_cache_to_file(self):
        with self._path.open(mode='w+') as f:
            for frame in self.__option_cache:
                f.write('option {0:08x}\n'.format(frame))
            for frame in self.__mixer_cache:
                f.write('mixer {0:08x}\n'.format(frame))
            for frame in self.__out_cache:
                f.write('out {0:08x}\n'.format(frame))

    def __load_settings(self):
        self.__load_option_settings()
        for target in self.get_mixer_labels():
            for src in self.get_mixer_src_labels():
                db = self.get_mixer_src(target, src)
                self.set_mixer_src(target, src, db)
        for target in self.get_out_labels():
            db = self.get_out_volume(target)
            self.set_out_volume(target, db)

    def get_model_name(self):
        return self.__name

    #
    # Configuration for options.
    #
    def __create_option_initial_cache(self):
        cache = [0x00] * 3
        self.__create_multiple_option_initial_cache(cache)
        self.__create_single_option_initial_cache(cache)
        return cache

    def __load_option_settings(self):
        # Assist ALSA fireface driver to handle MIDI messages from Fireface 400.
        if self.__name == 'Fireface400':
            FFOptionReg.build_single_option(self.__option_cache,
                                            'midi-low-addr', '0x00000000', True)
        req = Hinawa.FwReq()
        frames = pack('<3I', *self.__option_cache)
        req.transaction(self.get_node(), Hinawa.FwTcode.WRITE_BLOCK_REQUEST,
                        self.__regs[0], len(frames), frames)

    def __create_multiple_option_initial_cache(self, cache):
        default_params = {
            'line-in':          '-10dB',
            'line-out':         '-10dB',
            'primary-clk-src':  FFClkLabels.SPDIF.value,
        }
        for target, value in default_params.items():
            FFOptionReg.build_multiple_option(cache, target, value)

    def get_multiple_option_labels(self):
        return FFOptionReg.get_multiple_option_labels()

    def get_multiple_option_value_labels(self, target):
        return FFOptionReg.get_multiple_option_value_labels(target)

    def set_multiple_option(self, target, val):
        FFOptionReg.build_multiple_option(self.__option_cache, target, val)
        self.__write_cache_to_file()
        self.__load_option_settings()

    def get_multiple_option(self, target):
        return FFOptionReg.parse_multiple_option(self.__option_cache, target)

    def __create_single_option_initial_cache(self, cache):
        default_params = {
            'auto-sync': {
                'internal':     True,   'base-441':     True,
                'base-480':     True,   'double':       True,
                'quadruple':    True,
            },
            'front-input': {
                'in-1':         False,  'in-7':         False,
                'in-8':         False,
            },
            'rear-input': {
                'in-1':         True,   'in-7':         True,
                'in-8':         True,
            },
            'phantom-power': {
                'mic-7':        False,  'mic-8':        False,
                'mic-9':        False,  'mic-10':       False,
            },
            'spdif-in': {
                'optical':      False,  'track-maker':  False,
            },
            'spdif-out': {
                'professional': False,  'emphasis':     False,
                'non-audio':    False,  'optical':      False,
            },
            'instruments': {
                'drive':        False,  'limit':        True,
                'speaker-emu':  False,
            },
            'wdclk-out': {
                'single-speed': True,
            },
            'in-error': {
                'continue':     True,
            },
        }
        for target, params in default_params.items():
            for item, enable in params.items():
                FFOptionReg.build_single_option(cache, target, item, enable)

    def get_single_option_labels(self):
        return FFOptionReg.get_single_option_labels()

    def get_single_option_item_labels(self, target):
        return FFOptionReg.get_single_option_item_labels(target)

    def set_single_option(self, target, item, enable):
        FFOptionReg.build_single_option(self.__option_cache, target, item,
                                        enable)
        self.__write_cache_to_file()
        self.__load_option_settings()

    def get_single_option(self, target, item):
        return FFOptionReg.parse_single_option(self.__option_cache, target,
                                               item)

    def get_sync_status(self):
        req = Hinawa.FwReq()
        frames = bytearray(8)
        frames = req.transaction(self.get_node(),
                    Hinawa.FwTcode.READ_BLOCK_REQUEST,
                    0x0000801c0000, 8, frames)
        quads = unpack('<2I', frames)

        return FFStatusReg.parse(quads)

    #
    # Configuration for internal multiplexer.
    #
    def __create_mixer_initial_cache(self):
        targets = self.get_mixer_labels()
        srcs = self.get_mixer_src_labels()
        cache = [0x00] * (len(targets) * 2 * self.__spec['avail'])
        for i, target in enumerate(targets):
            for src in srcs:
                # Supply diagonal stream sources to each mixers.
                if src != 'stream-{0}'.format(i + 1):
                    val = self.__MUTE_VAL
                else:
                    val = self.__ZERO_VAL
                offset = FFMixerRegs.calculate_src_offset(self.__spec, target,
                                                          src)
                cache[offset // 4] = val
        return cache

    def get_mixer_labels(self):
        return FFMixerRegs.get_mixer_labels(self.__spec)

    def get_mixer_src_labels(self):
        return FFMixerRegs.get_mixer_src_labels(self.__spec)

    def get_mixer_mute_db(self):
        return self.get_db_mute()

    def get_mixer_min_db(self):
        return self.get_db_min()

    def get_mixer_max_db(self):
        return self.get_db_max()

    def set_mixer_src(self, target, src, db):
        offset = FFMixerRegs.calculate_src_offset(self.__spec, target, src)
        val = self.__build_val_from_db(db)
        data = pack('<I', val)
        req = Hinawa.FwReq()
        req.transaction(self.get_node(), Hinawa.FwTcode.WRITE_BLOCK_REQUEST,
                        self.__regs[1] + offset, len(data), data)
        self.__mixer_cache[offset // 4] = val
        self.__write_cache_to_file()

    def get_mixer_src(self, target, src):
        offset = FFMixerRegs.calculate_src_offset(self.__spec, target, src)
        return self.__parse_val_to_db(self.__mixer_cache[offset // 4])

    #
    # Configuration for output.
    #
    def __create_out_initial_cache(self):
        targets = FFOutRegs.get_out_labels(self.__spec)
        cache = [0x00] * len(targets)
        for target in targets:
            offset = FFOutRegs.calculate_out_offset(self.__spec, target)
            cache[offset // 4] = self.__ZERO_VAL
        return cache

    def get_out_labels(self):
        return FFOutRegs.get_out_labels(self.__spec)

    def set_out_volume(self, target, db):
        if db > self.get_db_max():
            raise ValueError('Invalid argument for db.')
        offset = FFOutRegs.calculate_out_offset(self.__spec, target)
        val = self.__build_val_from_db(db)
        data = pack('<I', val)
        req = Hinawa.FwReq()
        req.transaction(self.get_node(), Hinawa.FwTcode.WRITE_BLOCK_REQUEST,
                        self.__regs[2] + offset, len(data), data)
        self.__out_cache[offset // 4] = val

    def get_out_volume(self, target):
        offset = FFOutRegs.calculate_out_offset(self.__spec, target)
        return self.__parse_val_to_db(self.__out_cache[offset // 4])

    #
    # Helper methods.
    #
    @staticmethod
    def __build_val_from_db(db: float):
        return int(0x8000 * pow(10, db / 20))

    @staticmethod
    def __parse_val_to_db(val: int):
        if val == 0:
            return float('-inf')
        return 20 * log10(val / 0x8000)

    @classmethod
    def get_db_mute(cls):
        return cls.__parse_val_to_db(cls.__MUTE_VAL)

    @classmethod
    def get_db_zero(cls):
        return cls.__parse_val_to_db(cls.__ZERO_VAL)

    @classmethod
    def get_db_min(cls):
        return cls.__parse_val_to_db(cls.__MIN_VAL)

    @classmethod
    def get_db_max(cls):
        return cls.__parse_val_to_db(cls.__MAX_VAL)



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from pathlib import Path
from json import load, dump

from hinawa_utils.bebob.bebob_unit import BebobUnit
from hinawa_utils.bebob.extensions import BcoPlugInfo
from hinawa_utils.ta1394.general import AvcConnection
from hinawa_utils.ta1394.ccm import AvcCcm

from hinawa_utils.bebob.apogee_protocol import (
    HwCmd, DisplayCmd, OptIfaceCmd, MicCmd, InputCmd, OutputCmd, MixerCmd,
    RouteCmd, KnobCmd, SpdifResampleCmd
)

__all__ = ['ApogeeEnsembleUnit']


class ApogeeEnsembleUnit(BebobUnit):
    __CLOCK_SRCS = {
        'Coaxial':      AvcCcm.get_unit_signal_addr('external', 4),
        'Optical':      AvcCcm.get_unit_signal_addr('external', 5),
        'Word-clock':   AvcCcm.get_unit_signal_addr('external', 6),
    }

    def __init__(self, path):
        super().__init__(path)

        if (self.vendor_id, self.model_id) != (0x0003db, 0x01eeee):
            raise OSError('Not supported.')

        guid = self.get_property('guid')
        self.__path = Path('/tmp/hinawa-{0:08x}'.format(guid))

        if self.__path.exists() and self.__path.is_file():
            self.__load_cache()
        else:
            self.__create_cache()
            self.__set_from_cache()
            self.__save_cache()

    def __load_cache(self):
        with self.__path.open(mode='r') as f:
            self.__cache = load(f)

    def __save_cache(self):
        with self.__path.open(mode='w+') as f:
            dump(self.__cache, f)

    def __create_cache(self):
        cache = {}
        HwCmd.create_cache(cache)
        DisplayCmd.create_cache(cache)
        OptIfaceCmd.create_cache(cache)
        MicCmd.create_cache(cache)
        InputCmd.create_cache(cache)
        OutputCmd.create_cache(cache)
        MixerCmd.create_cache(cache)
        RouteCmd.create_cache(cache)
        SpdifResampleCmd.create_cache(cache)
        self.__cache = cache

    def __set_from_cache(self):
        val = HwCmd.get_cd_mode(self.__cache)
        HwCmd.set_cd_mode(self.__cache, self.fcp, val)
        val = HwCmd.get_16bit_mode(self.__cache)
        HwCmd.set_16bit_mode(self.__cache, self.fcp, val)

        val = DisplayCmd.get_illuminate(self.__cache)
        DisplayCmd.set_illuminate(self.__cache, self.fcp, val)
        val = DisplayCmd.get_mode(self.__cache)
        DisplayCmd.set_mode(self.__cache, self.fcp, val)
        val = DisplayCmd.get_target(self.__cache)
        DisplayCmd.set_target(self.__cache, self.fcp, val)
        val = DisplayCmd.get_overhold(self.__cache)
        DisplayCmd.set_overhold(self.__cache, self.fcp, val)

        for target in OptIfaceCmd.get_target_labels():
            val = OptIfaceCmd.get_mode(self.__cache, target)
            OptIfaceCmd.set_mode(self.__cache, self.fcp, target, val)

        for target in MicCmd.get_mic_labels():
            val = MicCmd.get_power(self.__cache, target)
            MicCmd.set_power(self.__cache, self.fcp, target, val)
            val = MicCmd.get_polarity(self.__cache, target)
            MicCmd.set_polarity(self.__cache, self.fcp, target, val)

        for target in InputCmd.get_in_labels():
            val = InputCmd.get_soft_limit(self.__cache, target)
            InputCmd.set_soft_limit(self.__cache, self.fcp, target, val)
            val = InputCmd.get_attr(self.__cache, target)
            InputCmd.set_attr(self.__cache, self.fcp, target, val)

        for target in OutputCmd.get_target_labels():
            val = OutputCmd.get_attr(self.__cache, target)
            OutputCmd.set_attr(self.__cache, self.fcp, target, val)

        for target in MixerCmd.get_target_labels():
            for src in MixerCmd.get_src_labels():
                vals = MixerCmd.get_src_gain(self.__cache, target, src)
                MixerCmd.set_src_gain(self.__cache, self.fcp, target, src,
                                      *vals)

        for target in RouteCmd.get_out_labels():
            src = RouteCmd.get_out_src(self.__cache, target)
            RouteCmd.set_out_src(self.__cache, self.fcp, target, src)
        for target in RouteCmd.get_cap_labels():
            src = RouteCmd.get_cap_src(self.__cache, target)
            RouteCmd.set_cap_src(self.__cache, self.fcp, target, src)
        for target in RouteCmd.get_hp_labels():
            src = RouteCmd.get_hp_src(self.__cache, target)
            RouteCmd.set_hp_src(self.__cache, self.fcp, target, src)

        params = SpdifResampleCmd.get_params(self.__cache)
        SpdifResampleCmd.set_params(self.__cache, self.fcp, *params)

    def __get_clock_plugs(self):
        plugs = {}
        info = AvcConnection.get_subunit_plug_info(self.fcp, 'music', 0)
        for direction in ('input', 'output'):
            for i in range(info[direction]):
                addr = BcoPlugInfo.get_subunit_addr(direction, 'music', 0, i)
                plug_type = BcoPlugInfo.get_plug_type(self.fcp, addr)
                if plug_type == 'Sync':
                    plugs[direction] = \
                        AvcCcm.get_subunit_signal_addr('music', 0, i)
                    break
            else:
                raise OSError('Unexpected state of device for clock source.')
        return plugs

    def get_clock_src_labels(self):
        labels = list(self.__CLOCK_SRCS.keys())
        labels.append('Internal')
        return labels

    def set_clock_src(self, src):
        if self.get_property('streaming'):
            raise OSError('Packet streaming started.')
        if src not in self.__CLOCK_SRCS and src != 'Internal':
            raise ValueError('Invalid argument for source of clock.')
        plugs = self.__get_clock_plugs()
        dst = plugs['input']
        if src == 'Internal':
            src = plugs['output']
        else:
            src = self.__CLOCK_SRCS[src]
        AvcCcm.set_signal_source(self.fcp, src, dst)

    def get_clock_src(self):
        plugs = self.__get_clock_plugs()
        dst = plugs['input']
        src = AvcCcm.get_signal_source(self.fcp, dst)
        addr = AvcCcm.parse_signal_addr(plugs['output'])
        if AvcCcm.compare_addrs(src, addr):
            return 'Internal'
        for name, addr in self.__CLOCK_SRCS.items():
            if AvcCcm.compare_addrs(src, AvcCcm.parse_signal_addr(addr)):
                return name
        raise OSError('Unexpected state of device.')

    def get_stream_mode_labels(self):
        return HwCmd.get_stream_mode_labels()

    def set_stream_mode(self, mode):
        HwCmd.set_stream_mode(mode)

    def get_stream_mode(self):
        sync_plug_ids = {
            5: '8x8',
            6: '10x10',
            7: '16x16',
        }
        plugs = self.__get_clock_plugs()
        addr = AvcCcm.parse_signal_addr(plugs['output'])
        plug_id = addr['data']['plug']
        if plug_id not in sync_plug_ids:
            raise OSError('Unexpected state of device.')
        return sync_plug_ids[plug_id]

    def reset_meters(self):
        DisplayCmd.reset_meter(self.fcp)

    def set_cd_mode(self, enable):
        if enable:
            if self.__cache['hw']['16bit'] != 'spdif-coax-out-1/2':
                raise ValueError('16bit-mode should be spdif-coax-out-1/2.')
            rate = AvcConnection.get_plug_signal_format(self.fcp, 'output', 0)
            if rate != 44100:
                raise ValueError('Sampling rate should be 44100.')
        HwCmd.set_cd_mode(self.__cache, self.fcp, enable)
        self.__save_cache()

    def get_cd_mode(self):
        return HwCmd.get_cd_mode(self.__cache)

    def get_16bit_mode_labels(self):
        return HwCmd.get_16bit_mode_labels()

    def set_16bit_mode(self, target):
        if target != 'none':
            rate = AvcConnection.get_plug_signal_format(self.fcp, 'output', 0)
            if rate not in (44100, 48000):
                raise ValueError('Sampling rate should be 44100 or 48000.')
        HwCmd.set_16bit_mode(self.__cache, self.fcp, target)
        self.__save_cache()

    def get_16bit_mode(self):
        return HwCmd.get_16bit_mode(self.__cache)

    # Hardware configurations.
    def set_display_mode(self, enable):
        DisplayCmd.set_mode(self.__cache, self.fcp, enable)
        self.__save_cache()

    def get_display_mode(self):
        return DisplayCmd.get_mode(self.__cache)

    def get_display_target_labels(self):
        return DisplayCmd.get_target_labels()

    def set_display_target(self, target):
        DisplayCmd.set_target(self.__cache, self.fcp, target)
        self.__save_cache()

    def get_display_target(self):
        return DisplayCmd.get_target(self.__cache)

    def set_display_illuminate(self, enable):
        DisplayCmd.set_illuminate(self.__cache, self.fcp, enable)
        self.__save_cache()

    def get_display_illuminate(self):
        return DisplayCmd.get_illuminate(self.__cache)

    def set_display_overhold(self, enable):
        DisplayCmd.set_overhold(self.__cache, self.fcp, enable)
        self.__save_cache()

    def get_display_overhold(self):
        return DisplayCmd.get_overhold(self.__cache)

    def get_opt_iface_target_labels(self):
        return OptIfaceCmd.get_target_labels()

    def get_opt_iface_mode_labels(self):
        return OptIfaceCmd.get_mode_labels()

    def set_opt_iface_mode(self, target, mode):
        OptIfaceCmd.set_mode(self.__cache, self.fcp, target, mode)
        self.__save_cache()

    def get_opt_iface_mode(self, target):
        return OptIfaceCmd.get_mode(self.__cache, target)

    # Knob configurations.
    def get_knob_out_labels(self):
        return KnobCmd.get_knob_out_labels()

    def set_knob_out_volume(self, target, db):
        KnobCmd.set_out_vol(self.fcp, target, db)

    def get_knob_states(self):
        return KnobCmd.get_states(self.fcp)

    # Microphone configurations.
    def get_mic_labels(self):
        return MicCmd.get_mic_labels()

    def set_polarity(self, target, invert):
        MicCmd.set_polarity(self.__cache, self.fcp, target, invert)
        self.__save_cache()

    def get_polarity(self, target):
        return MicCmd.get_polarity(self.__cache, target)

    def set_phantom_power(self, target, enable):
        MicCmd.set_power(self.__cache, self.fcp, target, enable)
        self.__save_cache()

    def get_phantom_power(self, target):
        return MicCmd.get_power(self.__cache, target)

    # Line input/output configurations.
    def get_line_in_labels(self):
        return InputCmd.get_in_labels()

    def set_soft_limit(self, target, enable):
        InputCmd.set_soft_limit(self.__cache, self.fcp, target, enable)
        self.__save_cache()

    def get_soft_limit(self, target):
        return InputCmd.get_soft_limit(self.__cache, target)

    def get_in_attr_labels(self):
        return InputCmd.get_attr_labels()

    def set_in_attr(self, target, attr):
        InputCmd.set_attr(self.__cache, self.fcp, target, attr)
        self.__save_cache()

    def get_in_attr(self, target):
        return InputCmd.get_attr(self.__cache, target)

    def get_line_out_labels(self):
        return OutputCmd.get_target_labels()

    def get_out_attr_labels(self):
        return OutputCmd.get_attr_labels()

    def set_out_attr(self, target, attr):
        OutputCmd.set_attr(self.__cache, self.fcp, target, attr)
        self.__save_cache()

    def get_out_attr(self, target):
        return OutputCmd.get_attr(self.__cache, target)

    # Route configurations.
    def get_out_labels(self):
        return RouteCmd.get_out_labels()

    def get_out_src_labels(self):
        return RouteCmd.get_out_src_labels()

    def set_out_src(self, target, src):
        RouteCmd.set_out_src(self.__cache, self.fcp, target, src)
        self.__save_cache()

    def get_out_src(self, target):
        return RouteCmd.get_out_src(self.__cache, target)

    def get_cap_labels(self):
        return RouteCmd.get_cap_labels()

    def get_cap_src_labels(self):
        return RouteCmd.get_cap_src_labels()

    def set_cap_src(self, target, src):
        RouteCmd.set_cap_src(self.__cache, self.fcp, target, src)
        self.__save_cache()

    def get_cap_src(self, target):
        return RouteCmd.get_cap_src(self.__cache, target)

    def get_hp_labels(self):
        return RouteCmd.get_hp_labels()

    def get_hp_src_labels(self):
        return RouteCmd.get_hp_src_labels()

    def set_hp_src(self, target, src):
        RouteCmd.set_hp_src(self.__cache, self.fcp, target, src)
        self.__save_cache()

    def get_hp_src(self, target):
        return RouteCmd.get_hp_src(self.__cache, target)

    # Internal multiplexer configuration.
    def get_mixer_labels(self):
        return MixerCmd.get_target_labels()

    def get_mixer_src_labels(self):
        return MixerCmd.get_src_labels()

    def set_mixer_src(self, target, src, db, balance):
        MixerCmd.set_src_gain(self.__cache, self.fcp, target, src, db, balance)
        self.__save_cache()

    def get_mixer_src(self, target, src):
        return MixerCmd.get_src_gain(self.__cache, target, src)

    # S/PDIF resampler configuration.
    def get_spdif_resample_iface_labels(self):
        return SpdifResampleCmd.get_iface_labels()

    def get_spdif_resample_direction_labels(self):
        return SpdifResampleCmd.get_direction_labels()

    def get_spdif_resample_rate_labels(self):
        return SpdifResampleCmd.get_rate_labels()

    def set_spdif_resample(self, enable, iface, direction, rate):
        SpdifResampleCmd.set_params(self.__cache, self.fcp, enable, iface,
                                    direction, rate)
        self.__save_cache()

    def get_spdif_resample(self):
        return SpdifResampleCmd.get_params(self.__cache)



# SPDX-License-Identifier: LGPL-3.0-or-later
# Copyright (C) 2018 Takashi Sakamoto

from hinawa_utils.bebob.phase_go_protocol_abstract import PhaseGoProtocolAbstract

from hinawa_utils.ta1394.audio import AvcAudio

__all__ = ['PhaseGoProtocolCoax']


class PhaseGoProtocolCoax(PhaseGoProtocolAbstract):
    __MIXER_OUTPUT_FB = 1
    __OUTPUT_LABELS = {
        'analog-1/2':       1,
        'headphone-1/2':    2,
        'digital-1/2':      3,
    }

    __ANALOG_INPUT_LEVELS = {
        'low':      bytearray((0xf4, 0x00)),
        'middle':   bytearray((0xfd, 0x00)),
        'high':     bytearray((0x00, 0x00)),
    }
    __ANALOG_INPUT_FB = 2
    __ANALOG_INPUT_FB_PN = 0

    def __init__(self, fcp):
        super().__init__(fcp, self.__MIXER_OUTPUT_FB, self.__OUTPUT_LABELS)

    def get_analog_input_level_labels(self):
        return list(self.__ANALOG_INPUT_LEVELS.keys())

    def set_analog_input_level(self, level):
        if level not in self.__ANALOG_INPUT_LEVELS:
            raise ValueError('Invalid argument for input level.')
        data = self.__ANALOG_INPUT_LEVELS[level]
        AvcAudio.set_feature_volume_state(self.fcp, 0, 'current',
                                          self.__ANALOG_INPUT_FB,
                                          self.__ANALOG_INPUT_FB_PN, data)

    def get_analog_input_level(self):
        result = AvcAudio.get_feature_volume_state(self.fcp, 0, 'current',
                                                   self.__ANALOG_INPUT_FB, self.__ANALOG_INPUT_FB_PN)
        for name, data in self.__ANALOG_INPUT_LEVELS.items():
            if data == result:
                return name
        else:
            raise OSError('Unexpected data for input level')
